//---------------------------------------------------------------------------
//
//  Module: vddctl.c
//
//  Purpose:
//     This module declares the virtual device (defines the Device
//     Declaration Block or DDB) and has the VDD_Control routine for
//     dispatching messages.
//
//     It also contains a number of the routines for handling the messages,
//     including VDD_Create_VM, VDD_Destroy_VM, VDD_VM_Critical_Init,
//     VDD_VM_Init, VDD_VM_Set_Device_Focus, VDD_Sys_Critical_Exit and
//     VDD_System_Exit.
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              3/25/93   BryanW      Rewritten in 'C'.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vddvga.h"

#pragma data_seg( "_LDATA", "CODE" ) 

//=============================================================================
//          V I R T U A L   D E V I C E   D E C L A R A T I O N
//=============================================================================

DDB VDD_DDB =
   { NULL,                      // reserved, must be NULL
     DDK_Version,               // reserved, DDK_Version
     VDD_Device_ID,             // Device ID
     VDD_Major_Ver,             // Major Version
     VDD_Minor_Ver,             // Minor Version
     NULL,
     "VDD",
     VDD_Init_Order,
     (DWORD) vmmwrapVxDControlProc,
     NULL,
     NULL,
     NULL,
     NULL,
     NULL,
     &VDD_Service_Table,
     VDD_Num_Services
   } ;

//=============================================================================
//           V M M   M E S S A G E   D I S P A T C H   T A B L E 
//=============================================================================

//
// This table is used by the vmmwrapVxDControlProc defined in VMMWRAP.ASM.                  
// It lists the messages and associated dispatch functions, it must be
// be terminated with -1 and NULL.                  
//

DISPATCHINFO alpVxDDispatchProcs[] =
   { Set_Device_Focus,    VDD_Set_Device_Focus,
     Begin_Message_Mode,  VDD_Ctl_Begin_Msg_Mode,
     End_Message_Mode,    VDD_Ctl_End_Msg_Mode,
     VM_Suspend,          VDD_Suspend_VM,
     VM_Resume,           VDD_Resume_VM,
     Create_VM,           VDD_Create_VM,
     Destroy_VM,          VDD_Destroy_VM,
     VM_Critical_Init,    VDD_VM_Critical_Init,
     VM_Init,             VDD_VM_Init,
     Sys_VM_Init,         VDD_Sys_VM_Init,
     System_Exit,         VDD_System_Exit,
     Device_Init,         VDD_Device_Init,
     Sys_Critical_Init,   VDD_Sys_Critical_Init,
     Init_Complete,       VDD_VMMem_InitComplete,
     Sys_Critical_Exit,   VDD_Sys_Critical_Exit,
#ifdef DEBUG
     Debug_Query,         VDD_Debug_Query
#endif
     -1,                  NULL } ;


//=============================================================================
//                    G L O B A L   D A T A   A R E A 
//=============================================================================

DWORD  dwVidCBOff, dwVidFlags, dwVTFlags ;

DWORD  hVidFocusVM, hVidCRTCVM, hVidMemCVM,
       hVidLatchesVM, hVidStartSRMemCVM ;

DWORD  hVidMsgPseudoVM ;
VDDCB  VidMsgPseudoCB ;

DWORD  dwVidScrollFreq ;

//=============================================================================
//                             L C O D E
//=============================================================================

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Create_VM
//  
//  Description:
//      Init data structures, allocate initial video memory save buffers.
//      Only error possible is failure to allocate memory.  Note that
//      Device_Init is the Create_VM for the SYSTEM VM.
//
//  Parameters:
//      DWORD hVM
//         VM handle
//  
//      PCRS_32 pCRS
//         pointer to client regs
//  
//  Return (BOOL):
//      TRUE if successful
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_Create_VM
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   PVDDCB  pVMCB ;

   UNUSED_PARAM( pCRS ) ;

   TRACEMSG( "VDD_Create_VM\r\n" ) ;

   pVMCB = (PVDDCB)(hVM + dwVidCBOff) ;

   pVMCB -> VDD_PIF = wPIFSave ;

   if (!VDD_VMMem_CreateVM( hVM, pVMCB ))
      return FALSE ;

   VDD_State_CreateVM( hVM, pVMCB ) ;

   return ( TRUE ) ;
   
} // VDD_Create_VM()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Sys_VM_Init
//  
//  Description:
//      Initialized the system VM.
//  
//  Parameters:
//      DWORD hVM
//         handle to system VM
//  
//      PCRS_32 pCRS
//         pointer to client regs
//  
//  Return (BOOL):
//      TRUE if successful.
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_Sys_VM_Init
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   VDD_VM_Init( hVM, pCRS ) ;
   dwVidFlags |= fVid_SysVMI ;

   return TRUE ;

} // VDD_Sys_VM_Init()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_VM_Init
//  
//  Description:
//     Initialize VM's memory, enable/disable I/O port trapping and
//     initialize device state by calling INT 10.
//  
//  Parameters:
//      DWORD hVM
//         handle to system VM
//  
//      PCRS_32 pCRS
//         pointer to client regs
//  
//  Return (BOOL):
//      TRUE if successful.
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_VM_Init
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   PVDDCB  pVMCB ;

   UNUSED_PARAM( pCRS ) ;

   pVMCB = (PVDDCB)(hVM + dwVidCBOff) ;

   VDD_State_VMInit( hVM, pVMCB ) ;

   return TRUE ;

} // VDD_VM_Init()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Set_Device_Focus
//  
//  Description:
//      Called when switching the display focus.
//  
//  Parameters:
//      DWORD hVM
//  
//      DWORD dwVID
//         VxD ID receiving focus
//  
//      DWORD dwFlags
//         flags
//  
//      DWORD hAssocVM
//         handle of associated VM
//  
//  Return (BOOL):
//      TRUE if successful
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_Set_Device_Focus
(
    DWORD           hVM,
    DWORD           dwVID,
    DWORD           dwFlags,
    DWORD           hAssocVM
)
{
   PVDDCB  pVMCB ;

   UNUSED_PARAM( dwFlags ) ;
   UNUSED_PARAM( hAssocVM ) ;

   pVMCB = (PVDDCB)(hVM + dwVidCBOff) ;

   // If critical set focus call or a specific VDD
   // set focus, then do the work...

   if ((!dwVID) || (VDD_Device_ID == dwVID))
   {
      if (hVM != hVidFocusVM)
      {
         hVidFocusVM = hVM ;

         if (0 == (dwVidFlags & fVid_MsgA))
         {
            if (dwVidFlags & fVid_Crit)
               dwVidFlags |= fVid_SFPending ;
            else
               VDD_Ctl_Set_Focus( hVM, pVMCB ) ;
         }
      }
   }

   return TRUE ;

} // VDD_Set_Device_Focus()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Ctl_Set_Focus
//  
//  Description:
//     Switches display focus.
//  
//  Parameters:
//      DWORD hVM
//         VM Handle of VM to switch display focus to
//  
//  Return (BOOL):
//     TRUE if successful
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_Ctl_Set_Focus
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   VDD_State_ScrnOff( hVM ) ;
   VDD_State_SetCRTCOwner( hVM, pVMCB ) ;
   VDD_VMMem_SetDeviceFocus( hVM ) ;
   VDD_State_SetMemCOwner( hVM, pVMCB ) ;
   VDD_State_ScrnOn( hVM ) ;

   return TRUE ;

} // VDD_Ctl_Set_Focus()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Ctl_End_Dsp_Crit
//  
//  Description:
//      handle a pending set focus, if necessary
//  
//  Parameters:
//      None.
//  
//  Return (BOOL):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Ctl_End_Dsp_Crit()
{

   if (dwVidFlags & fVid_SFPending)
   {
      DWORD   hVM ;
      PVDDCB  pVMCB ;

      dwVidFlags &= ~fVid_SFPending ;

      // set focus...

      hVM = hVidFocusVM ;
      pVMCB = (PVDDCB)(hVM + dwVidCBOff) ;

      if (0 == (dwVidFlags & fVid_MsgA))
      {
         if (dwVidFlags & fVid_Crit)
            dwVidFlags |= fVid_SFPending ;
         else
            VDD_Ctl_Set_Focus( hVM, pVMCB ) ;
      }
   }

} // VDD_Ctl_End_Dsp_Crit()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Destroy_VM
//  
//  Description:
//      Deallocate pages used for video RAM
//  
//  Parameters:
//      DWORD hVM
//         handle of VM being destroyed
//  
//      PCRS_32 pCRS
//         client reg structure
//  
//  Return (BOOL):
//      TRUE on success
//
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_Destroy_VM
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   PVDDCB  pVMCB ;

   UNUSED_PARAM( pCRS ) ;

   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;
   VDD_State_DestroyVM( hVM, pVMCB ) ;
   VDD_VMMem_DestroyVM( hVM ) ;

   return TRUE ;

} // VDD_Destroy_VM()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Suspend_VM
//  
//  Description:
//  
//  
//  Parameters:
//      DWORD hVM
//          handle of VM being suspended
//  
//      PCRS_32 pCRS
//          client reg structure
//  
//  Return (BOOL):
//      TRUE on success
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_Suspend_VM
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   PVDDCB  pVMCB ;

   UNUSED_PARAM( pCRS ) ;

   pVMCB = (PVDDCB)(hVM + dwVidCBOff) ;

   VDD_Proc_VMSuspend( hVM, pVMCB ) ;
   VDD_VMMem_VMSuspend( pVMCB ) ;

   return TRUE ;

} // VDD_Suspend_VM()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Resume_VM
//  
//  Description:
//  
//  
//  Parameters:
//      DWORD hVM
//          handle of VM being resumed
//  
//      PCRS_32 pCRS
//          client reg structure
//  
//  Return (BOOL):
//      FALSE if cannot resume
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_Resume_VM
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   PVDDCB pVMCB ;

   UNUSED_PARAM( pCRS ) ;

   pVMCB = (PVDDCB)(hVM + dwVidCBOff) ;

   if (!VDD_VMMem_VMResume( pVMCB ))
      return FALSE ;

   VDD_Proc_VMResume( hVM, pVMCB ) ;

   return TRUE ;

} // VDD_Resume_VM()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_System_Exit
//  
//  Description:
//      Clean up VM video state.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PCRS_32 pCRS
//         pointer to client regs
//  
//  Return (BOOL):
//      TRUE if successful
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_System_Exit
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   PVDDCB  pVMCB ;

   UNUSED_PARAM( pCRS ) ;

   pVMCB = (PVDDCB)(hVM + dwVidCBOff) ;

   VDD_State_SystemExit( hVM, pVMCB ) ;

   return TRUE ;

} // VDD_System_Exit()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Sys_Critical_Exit
//  
//  Description:
//      Restore original VM state
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//  Return (VOID):
//      nothing.
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Sys_Critical_Exit
(
    DWORD           hVM
)
{
   UNUSED_PARAM( hVM ) ;

   VDD_Int_SysCriticalExit() ;

} // VDD_Sys_Critical_Exit()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_VM_Critical_Init
//  
//  Description:
//      Finish setting up VM for execution.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PCRS_32 pCRS
//         client regs
//  
//  Return (BOOL):
//     TRUE if successful
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_VM_Critical_Init
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   PVDDCB  pVMCB ;

   UNUSED_PARAM( pCRS ) ;

   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;
   VDD_VMMem_VMCriticalInit( hVM, pVMCB ) ;

   return ( TRUE ) ;

} // VDD_VM_Critical_Init()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Ctl_Begin_Msg_Mode
//  
//  Description:
//      Thunk to VDD_Begin_Msg_Mode service.
//  
//  Parameters:
//      DWORD hVM
//          handle to VM entering message mode.
//  
//      PCRS_32 pCRS
//          pointer to client reg structure.
//  
//  Return (BOOL):
//      TRUE on success.
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_Ctl_Begin_Msg_Mode
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   REGS  regs ;

   UNUSED_PARAM( pCRS ) ;

   regs.ebx = hVM ;
   VDD_Begin_Msg_Mode( &regs ) ;

   return TRUE ;

} // VDD_Ctl_Begin_Msg_Mode()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Ctl_End_Msg_Mode
//  
//  Description:
//      Ends message mode processing.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM leaving message mode
//  
//      PCRS_32 pCRS
//         pointer to client reg structure
//  
//  Return (BOOL):
//      TRUE on success.
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_Ctl_End_Msg_Mode
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   REGS  regs ;

   UNUSED_PARAM( pCRS ) ;

   regs.ebx = hVM ;

   VDD_End_Msg_Mode( &regs ) ;

   return TRUE ;

} // VDD_Ctl_End_Msg_Mode()

//---------------------------------------------------------------------------
//  End of File: vddctl.c
//---------------------------------------------------------------------------

