//---------------------------------------------------------------------------
//
//  Module: vddfonts.c
//
//  Purpose:
//     Text mode font handling routines.
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              4/24/93   BryanW      Rewritten in 'C'.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vddvga.h"
#include "vddphmem.h"

#pragma data_seg( "_LDATA", "CODE" )

//=============================================================================
//                    G L O B A L   D A T A   A R E A 
//=============================================================================

BYTE   bVidCharHgt ;
DWORD  dwVidMsgFontH = 0, dwVidMsgFont ;

//=============================================================================
//                             L C O D E
//=============================================================================

//--------------------------------------------------------------------------
//  
//  VOID VDD_Font_SaveFont
//  
//  Description:
//      Copy a 4Kb page of a text mode RAM font from physical
//      video memory to a page of VM save memory.  This routine
//      requires that the hardware be set up for the font read.
//      It handles finding a free spot in the fontmap table in
//      VDD_MStt, if not already assigned.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM's VDD control block
//  
//      PMSSPLANEINFO pMPI
//         pointer to plane info
//  
//      DWORD hPage
//         handle to page
//  
//      UINT uPhysPage
//         physical page
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Font_SaveFont
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPage
)
{
   BOOL   fSecondPage ;
   DWORD  dwFontIdx, dwFontPg ;

   dwFontIdx = ((PPAGEDATA) hPage) -> PDS_virtual_id ;
   fSecondPage = (dwFontIdx & 1) ;
   dwFontPg = dwFontIdx ;
   if (0 != (dwFontIdx >>= 1))
   {
      if (dwFontPg = pVMCB -> VDD_MStt.MSS_fontmap[ dwFontIdx ])
         dwFontPg += fSecondPage ;
      else
      {
         BYTE   bFontPg, bMapBits = 0xFE ;
         PSTR   pMap ;
         UINT   i ;

         // locate free position for font in save memory

         pMap = &pVMCB -> VDD_MStt.MSS_fontmap[ 1 ] ;

         // clear the allocated pages...

         for (i = 0; i < 7; i++)
         {
            bFontPg = *pMap++ ;
            bMapBits &= ~(1 << (bFontPg >> 1)) ;
         }

         // find empty slot

         for (i = 0; i < 7; i++)
            if (bMapBits & (i << 1))
               break ;

         dwFontPg = i << 1 ;
         pVMCB -> VDD_MStt.MSS_fontmap[ dwFontIdx ] = (BYTE) dwFontPg ;
         dwFontPg += fSecondPage ;
      }
   }

   if (VDD_VMMem_LockSaveMemPage( hVM, pMPI, dwFontPg ))
   {
      UINT   i ;
      PBYTE  pDst, pSrc ;

      pDst = (PBYTE) (dwFontPg << 12) + pMPI -> MSS_plane_addr ;
      pSrc = (PBYTE) (uPhysPage << 12) - 0xA0000 + dwVidPhysA0000 ;
      for (i = 0; i < P_SIZE; i++)
         *pDst++ = *pSrc++ ;
      vmmPageUnLock( pMPI -> MSS_plane_handle, 1, dwFontPg, 0 ) ;
   }

} // VDD_Font_SaveFont()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Font_RestoreFont
//  
//  Description:
//      Copy a 4Kb page of a text mode RAM font from VM save
//      memory to a physical video memory page.  This routine
//      requires that the hardware be set up for the font write.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM's VDD control block
//  
//      PMSSPLANEINFO pMPI
//         pointer to plane info
//  
//      DWORD hPage
//         handle to page
//  
//      UINT uPhysPage
//         physical page
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Font_RestoreFont
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPage
)
{

   BOOL   fSecondPage ;
   DWORD  dwFontIdx, dwFontPg ;

   dwFontIdx = ((PPAGEDATA) hPage) -> PDS_virtual_id ;
   fSecondPage = (dwFontIdx & 1) ;
   dwFontPg = dwFontIdx ;
   if (0 != (dwFontIdx >>= 1))
   {
      if (dwFontPg = pVMCB -> VDD_MStt.MSS_fontmap[ dwFontIdx ])
         dwFontPg += fSecondPage ;
      else
         VDD_VMMem_ZeroPage( uPhysPage ) ;
   }

   if (VDD_VMMem_LockSaveMemPage( hVM, pMPI, dwFontPg ))
   {
      UINT   i ;
      PBYTE  pDst, pSrc ;

      pSrc = (PBYTE) (dwFontPg << 12) + pMPI -> MSS_plane_addr ;
      pDst = (PBYTE) (uPhysPage << 12) - 0xA0000 + dwVidPhysA0000 ;
      for (i = 0; i < P_SIZE; i++)
         *pDst++ = *pSrc++ ;
      vmmPageUnLock( pMPI -> MSS_plane_handle, 1, dwFontPg, 0 ) ;
   }
   else
      VDD_VMMem_ZeroPage( uPhysPage ) ;

} // VDD_Font_RestoreFont()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Font_RestoreMsgFont
//  
//  Description:
//     Restore packed msg mode font into the first 2 pages of
//     video memory.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (BOOL):
//     FALSE if restore failed.
//
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_Font_RestoreMsgFont
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   UNUSED_PARAM( hVM ) ;

   VDD_State_SetMemCPlanar() ;
   VDD_VMMem_SelectPlane( 2, 4 ) ;

   VDD_Font_ExpandMsgFont( (PSTR) dwVidPhysA0000 ) ;

   return TRUE ;

} // VDD_Font_RestoreMsgFont()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Font_ExpandMsgFont
//  
//  Description:
//     Copy packed message mode font into the expanded 32 byte
//     per character form used in video memory.
//  
//  Parameters:
//      PSTR pDst
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Font_ExpandMsgFont
(
    PSTR            pDst
)
{
   UINT  i, j, uCellHeight, uCellFill ;
   PSTR  pSrc ;

   uCellHeight = (UINT) bVidCharHgt ;
   uCellFill = 32 - uCellHeight ;
   pSrc = (PSTR) dwVidMsgFont ;
   for (i = 0; i < 256; i++)
   {
      for (j = 0; j < uCellHeight; j++)
         *pDst++ = *pSrc++ ;
      for (j = 0; j < uCellFill; j++)
         *pDst++ = 0 ;
   }

} // VDD_Font_ExpandMsgFont()

//---------------------------------------------------------------------------
//  End of File: vddfonts.c
//---------------------------------------------------------------------------

