//---------------------------------------------------------------------------
//
//  Module: vddproc.c
//
//  Purpose:
//     This module provides most of routines called by VMM. The VDD_Control
//     procedure dispatches these VMM calls. VDD_VMCreate is called
//     when a VM is being created.  VDD_VMInit is called just before a VM
//     is run for the first time. VDD_VMSetFocus and is called when the
//     display focus is changed. Initialization is done through the
//     VDD_Sys_Critical_Init and VDD_Device_Init.
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//             10/20/93   BryanW      Rewritten in 'C'.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vddvga.h"

#pragma data_seg( "_LDATA", "CODE" ) 

//=============================================================================
//                    G L O B A L   D A T A   A R E A 
//=============================================================================

//---
//
// Delay window update by VDD_Mode_Update_Time to avoid screen
// size bounce during a 43 line mode change.
// This is related to the fVDD_ModeChange in VM's VDD_Flags.
// At the beginning of a mode change this bit is set,
// at the end of a mode change we set a long time out.  The long timeout
// gives the VM enough time to modify the state further (like change the
// font for 43 line mode) before we do a windowed VM update.
// When the timeout occurs or the VM goes idle, fModeChange is cleared and
// all subsequent timeouts are normal duration.
// 
// when fVDD_ModeChange is set - vdd_get_mod returns no changes,
//                             - vdd_VMIdle skips setting a timeout 
// 	                         - VDD_Set_VM_TimeOut sets a long timeout
//
//---

#define VDD_ModeChangeDelay   100

DWORD dwVidWinUpdateTime = 50 ;
DWORD dwVidModeUpdateTime = VDD_ModeChangeDelay ;
DWORD dwVidScrollFreq = 2 ;

// next procedure in release time slice chain

DWORD  dwVidVMIdleChain = 0 ;

//=============================================================================
//                             L C O D E
//=============================================================================

//--------------------------------------------------------------------------
//  
//  VOID VDD_VMIdle
//  
//  Description:
//      Check for screen update while VM is idle
//
//      Display changes are checked for after a VM has been executing for
//      a fixed period of time, or the VM has become idle(e.g. waiting for
//      keyboard input) in which case the VM will not be executing at
//      all. This is the routine that gets called when the VM is idle.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VMIdle
(
    DWORD           hVM
)
{
   PVDDCB  pVMCB ;
   
   pVMCB = (PVDDCB)(hVM + dwCBOff) ;

   if (pVMCB -> VDD_Flags & fVDDWin)
   {
#ifdef DEBUG
      if (pVMCB -> VDD_Flags & fVDD_ModeSet)
         TRACEMSG( "IDLE during video mode change!!\r\n" ) ;
#endif
      VDD_ClrVMTimeOut( hVM, pVMCB ) ;
      pVMCB -> VDD_Flags &= ~fVDD_ModeChange ;

      if (pVMCB -> VDD_Flags & fVDD_Upd)
         pVMCB -> VDD_Flags |= fVDD_IdleUpdate ;
      else
         VDD_UpdateWindow( hVM, pVMCB ) ;
   }

} // VDD_VMIdle()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VMTimeOut
//  
//  Description:
//      Display changes are checked for after a VM has been executing for
//      a fixed period of time, or the VM has become idle(e.g. waiting for
//      keyboard input) in which case the VM will not be executing much at
//      all. This is the routine that gets called when the VM has executed
//      for awhile
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PCRS_32 pCRS
//         pointer to client reg struct
//  
//      DWORD dwRefData
//         reference data
//  
//      DWORD dwExtra
//         number of ms elapsed since timeout was scheduled
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_VMTimeOut
(
    DWORD           hVM,
    PCRS_32         pCRS,
    DWORD           dwRefData,
    DWORD           dwExtra
)
{
   PVDDCB pVMCB ;

   pVMCB = (PVDDCB) dwRefData ;

   pVMCB -> VDD_Time_Out_Handle = 0 ;

   if (pVMCB -> VDD_Flags & fVDD_ModeSet)
      return ;

   pVMCB -> VDD_Flags &= ~fVDD_ModeChange ;

   if (pVMCB -> VDD_Flags & fVDD_Win)
   {
      UINT uPostponed ;

      uPostponed = pVMCB -> VDD_Postponed_TimeOuts ;
      pVMCB -> VDD_Postponed_TimeOuts = 0 ;

      if (0 == (pVMCB -> VDD_Flags & fVDD_PageFault))
      {
         VDD_Update_Window( hVM, pVMCB ) ;
         return ;
      }
      pVMCB -> VDD_Flags &= ~fVDD_PageFault ;
      if (uPostponed >= 8)
      {
         VDD_UpdateWindow( hVM, pVMCB ) ;
         return ;
      }
      pVMCB -> VDD_Postponed_TimeOuts++ ;
      VDD_SetVMTimeOut( hVM, pVMCB ) ;
   }

} // VDD_VMTimeOut()

//--------------------------------------------------------------------------
//  
//  VOID VDD_UpdateWindow
//  
//  Description:
//      Check for screen update
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/30/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_UpdateWindow
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   pVMCB -> VDD_Recent_User_Action = 0 ;

   if (pVMCB -> VDD_Flags & fVDD_Upd)
      return ;

   if (VDD_State_ChangeQuery( hVM, pVMCB ) ||
       VDD_Grab_ChangeQuery( hVM, pVMCB ))
   {
      pVMCB -> VDD_Flags |= fVDD_Upd ;
      pVMCB -> VDD_Update_Hit_Cnt++ ;
      if (pVMCB -> VDD_Stt.V_BiosMode < 0x0D)
         pVMCB -> VDD_Upd_Start = vmmGetVMExecTime( vmmGetSysVMHandle() ) ;
      vmmCallPriorityVMEvent( hVM, Low_Pri_Device_Boost,
                              PEF_Wait_For_STI | PEF_Wait_Not_Crit |
                              PEF_Always_Sched, (DWORD) pVMCB,
                              pVDD_Update_Event ) ;
   }
   else
   {
      pVMCB -> VDD_Update_Hit_Cnt = 0 ;
      VDD_SetVMTimeOut( hVM, pVMCB ) ;
   }

} // VDD_UpdateWindow()

//--------------------------------------------------------------------------
//  
//  VOID VDD_SetVMTimeOut
//  
//  Description:
//      Sets a timeout for specified VM
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/30/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_SetVMTimeOut
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if (pVMCB -> VDD_Flags & fVDD_Win)
   {
      if (0 == (pVMCB -> VDD_Time_Out_Handle))
      {
         DWORD dwTime ;

         dwTime = dwVidModeUpdateTime ;
         if (pVMCB -> VDD_Flags & fVDD_ModeChange)
         {
            dwTime = dwVidWinUpdateTime ;
            if (pVMCB -> VDD_Update_Hit_Cnt >= 3)
               dwTime <<= 2 ;
         }
         pVMCB -> VDD_Time_Out_Handle =
            vmmSetVMTimeOut( hVM, dwTime, pVDD_VMTimeOut, (DWORD) pVMCB ) ;

      }
   }

} // VDD_SetVMTimeOut()

//--------------------------------------------------------------------------
//  
//  VOID VDD_ClrVMTimeOut
//  
//  Description:
//      Clear any pending video update timeout for VM
//  
//  Parameters:
//      DWORD hVM
//         handle to VM 
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_ClrVMTimeOut
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   DWORD dwTO ;

   UNUSED_PARAM( hVM ) ;

   dwTO = pVMCB -> VDD_Time_Out_Handle ;
   pVMCB -> VDD_Time_Out_Handle = 0 ;
   vmmCancelTimeOut( dwTO ) ;

} // VDD_ClrVMTimeOut()

//--------------------------------------------------------------------------
//  
//  VOID VDD_UpdateEvent
//  
//  Description:
//      Call Shell_Event with update event for VMDOSAPP.	This
//      should only be called when critical section is unowned.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to system VM
//  
//      PCRS_32 pCRS
//         pointer to client reg structure
//  
//      DWORD dwRefData
//         reference data
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/30/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_UpdateEvent
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           dwRefData
)
{
   DWORD   hVM, dwFlags ;
   PVDDCB  pVMCB ;

   hVM = dwRefData - dwVidCBOff ;
   pVMCB = (PVDDCB) dwRefData ;

   if (vmmValidateVMHandle( hVM ))
   {
      dwFlags = VDA_DE_DispChngMin ;
      if (pVMCB -> VDD_Flags & fVDD_BoostUpdate)
      {
         pVMCB -> VDD_Flags &= ~fVDD_BoostUpdate ;
         dwFlags |= SE_WP_DispUpdBoost ;
      }

      if (!shellEvent( hVM, VDA_Display_Event, dwFlags,
                       pVDD_Update_CallBack, (DWORD) pVMCB ))
      {
         DEBUGMSG( "VDD: failed to post message for grabber\r\n" ) ;
         VDD_Update_CallBack( NULL, (DWORD) pVMCB ) ;
      }
   }

} // VDD_UpdateEvent()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Update_CallBack
//  
//  Description:
//      This routine is called when update event is completed
//  
//  Parameters:
//      PCRS_32 pCRS
//          pointer to client reg struct
//  
//      DWORD dwRefData
//          reference data
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/30/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Update_CallBack
(
    PCRS_32         pCRS,
    DWORD           dwRefData
)
{
   DWORD   hVM ;
   PVDDCB  pVMCB ;

   pVMCB = (PVDDCB) dwRefData ;

   pVMCB -> VDD_Flags &= ~fVDD_Upd ;
   hVM = dwRefData - dwVidCBOff ;

   if (vmmValidateVMHandle( hVM ))
   {
      if (pVMCB -> VDD_Flags & fVDD_IdleUpdate)
      {
         pVMCB -> VDD_Flags &= ~fVDD_IdleUpdate ;
         VDD_UpdateWindow( hVM, pVMCB ) ;

      }
      else
         VDD_SetVMTimeOut( hVM, pVMCB ) ;
   }

} // VDD_Update_CallBack()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VMDA_Grab
//  
//  Description:
//      If no Grab event is already queued, set up event to call
//      Shell_Event with grab event for VMDOSAPP.
//
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VMDA_Grab
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if (0 == (pVMCB -> VDD_Flags & fVDD_Grab))
   {
      pVMCB -> VDD_Flags |= fVDD_Grab ;

      vmmCallPriorityVMEvent( vmmGetSysVMHandle(), Low_Pri_Device_Boost,
                              PEF_Wait_For_STI | PEF_Wait_Not_Crit |
                              PEFT_Always_Sched, (DWORD) pVMCB,
                              pVDD_Grab_Event ) ;

   }

} // VDD_VMDA_Grab()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Grab_Event
//  
//  Description:
//      call Shell_Event with grab event for VMDOSAPP.
//  
//  Parameters:
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Grab_Event
(
    PVDDCB          pVMCB
)
{
   DWORD  hVM ;

   hVM = ((DWORD) pVMCB) - dwVidCBOff ;

   if (vmmValidateVMHandle( hVM ))
   {
      if (!shellEvent( hVM, VDA_Display_Event,
                       VDA_DE_ScreenGrabMin | SE_WP_PrtScBoost,
                       pVDD_Grab_CallBack, (DWORD) pVMCB ))
      {
         DEBUGMSG( "VDD: failed to post message for grabber\r\n" ) ;
         VDD_Grab_CallBack( NULL, (DWORD) pVMCB ) ;
      }
   }

} // VDD_Grab_Event()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Grab_CallBack
//  
//  Description:
//      Finished with grab, delete memory
//  
//  Parameters:
//      PCRS_32 pCRS
//         pointer to a WINOLDAPP client reg struct frame
//  
//      DWORD dwRefData
//         contains pointer to VM control block
//  
//  Return (VOID):
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Grab_CallBack
(
    PCRS_32         pCRS,
    DWORD           dwRefData
)
{
   DWORD   hVM ;
   PVDDCB  pVMCB ;

   UNUSED_PARAM( pCRS ) ;

   pVMCB = (PVDDCB) dwRefData ;
   hVM = dwRefData - dwVidCBOff ;

   pVMCB -> VDD_Flags &= ~fVDD_Grab ;
   VDD_Grab_DCopy( hVM, pVMCB ) ;

} // VDD_Grab_CallBack()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_VMResume
//  
//  Description:
//  
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Proc_VMResume
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if (pVMCB -> VDD_Flags & fVDD_CanRestore)
   {
      if (pVMCB -> VDD_Flags & fVDD_WaitAtt)
      {
         pVMCB -> VDD_Flags &= ~fVDD_WaitAtt ;

         if (0 == pVMCB -> VDD_NotifyEvent)
            pVMCB -> VDD_NotifyEvent =
               vmmScheduleVMEvent( hVM, pVDD_Proc_FGrndINT2F, 0 ) ;
      }
   }

} // VDD_Proc_VMResume()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_VMSuspend
//  
//  Description:
//  
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM's control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Proc_VMSuspend
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if (pVMCB -> VDD_Flags & fVDD_NoSaveRes)
      VDD_Proc_BGrndNotify( hVM, pVMCB ) ;

} // VDD_Proc_VMSuspend()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_BGrndINT2F
//  
//  Description:
//      Simulate INT 2F, AX=4001h into VM and AX=4002h if necessary.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      DWORD dwRefData
//         not used, here for callback compatibility
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Proc_BGrndINT2F
(
    DWORD           hVM,
    DWORD           dwRefData
)
{
   CRS_32   CRS ;
   PCRS_32  pCRS ;
   PVDDCB   pVMCB ;

   UNUSED_PARAM( dwRefData ) ;

   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;

   pVMCB -> VDD_NotifyEvent = 0 ;

   // reset VM's controller state to what was saved when
   // when the VM notified us that it could restore it's
   // own screen.

   if (pVMCB -> VDD_SavedStt)
   {
      pVMCB -> VDD_Stt =
         *((PREGSTATE)(pVMCB -> VDD_SavedStt)) ;

#ifdef DEBUG
      pVMCB -> VDD_Flags |= fVDD_NeedRestore ;
#endif
   }

   pCRS = (PCRS_32) (((PVMMCB) hVM) -> CB_Client_Pointer) ;

   vmmSaveClientState( hVM, pCRS, &CRS ) ;
   vmmBeginNestExec( hVM, pCRS ) ;

   ((PCRS_16) pCRS) -> Client_AX = 0x4001 ;
   vmmExecInt( hVM, pCRS, 0x2F ) ;

   if (0 == (pVMCB -> VDD_Flags & fVDD_WaitAtt))
   {
      // notify foreground too...

      ((PCRS_16) pCRS) -> Client_AX = 0x4002 ;
      pVMCB -> VDD_Flags |= fVDD_InInt2F ;
      vmmExecInt( hVM, pCRS, 0x2F ) ;
      pVMCB -> VDD_Flags &= ~fVDD_InInt2F ;
   }

   vmmEndNestExec( hVM, pCRS ) ;
   vmmRestoreClientState( hVM, pCRS, &CRS ) ;

} // VDD_Proc_BGrndINT2F()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_FGrndINT2F
//  
//  Description:
//      Simulate INT 2F, AX=4002h into VM
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      DWORD dwRefData
//         not used, here for callback compatibility
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Proc_FGrndINT2F
(
    DWORD           hVM,
    DWORD           dwRefData
)
{
   CRS_32   CRS ;
   PCRS_32  pCRS ;
   PVDDCB   pVMCB ;

   UNUSED_PARAM( dwRefData ) ;

   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;

   pVMCB -> VDD_NotifyEvent = 0 ;

   if (pVMCB -> VDD_Flags & fVDD_WaitAtt)
      return ;

   pCRS = (PCRS_32) (((PVMMCB) hVM) -> CB_Client_Pointer) ;

   vmmSaveClientState( hVM, pCRS, &CRS ) ;
   vmmBeginNestExec( hVM, pCRS ) ;

   ((PCRS_16) pCRS) -> Client_AX = 0x4002 ;
   pVMCB -> VDD_Flags |= fVDD_InInt2F ;
   vmmExecInt( hVM, pCRS, 0x2F ) ;
   pVMCB -> VDD_Flags &= ~fVDD_InInt2F ;

   vmmEndNestExec( hVM, pCRS ) ;
   vmmRestoreClientState( hVM, pCRS, &CRS ) ;

} // VDD_Proc_FGrndINT2F()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_BGrndNotify
//  
//  Description:
//      Schedule/Call VDD_Proc_BGrnd_INT2F
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Proc_BGrndNotify
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if (pVMCB -> VDD_NotifyEvent)
   {
      if (0 == (pVMCB -> VDD_Flags & fVDD_WaitAtt))
      {
         dwVidFlags |= fVid_VMsSuspended ;
         pVMCB -> VDD_Flags |= fVDD_WaitAtt ;
      }
      return ;
   }

   dwVidFlags |= fVid_VMsSuspended ;
   pVMCB -> VDD_Flags |= fVDD_WaitAtt ;
   if (vmmTestCurVMHandle( hVM ) &&
       (0 == (((PVMMCB) hVM) -> CB_VM_Status & VMStat_Suspended)))
      VDD_Proc_BGrndINT2F( hVM, 0 ) ;
   else
      pVMCB -> VDD_NotifyEvent =
         vmmScheduleVMEvent( hVM, pVDD_Proc_BGrndINT2F, 0 ) ;

} // VDD_Proc_BGrndNotify()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_NotifyBkgndVMs
//  
//  Description:
//      Do bkgnd notifications for background VM's that have
//      done the INT 2F to tell us that they know how to restore
//      their screen, if auto save/restore is not enabled.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM's control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Proc_NotifyBkgndVMs
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   DWORD  hCurVM ;

   if (0 == (pVMCB -> VDD_TFlags & (fVT_NoSaveResSysVM |
                                    fVT_NoSaveResOthers)))
      return ;

   for (hCurVM = vmmGetNextVMHandle( hVM );
        hCurVM != hVM;
        hCurVM = vmmGetNextVMHandle( hCurVM ))
   {

#ifdef SysVMin2ndBank
      // If SysVM is running in 2nd bank, we don't need
      // to notify it because it never loses it's memory

      if (pVMCB -> VDD_TFlags & fVT_SysVMnot1stBank)
         if (vmmTestSysVMHandle( hCurVM ))
            continue ;
#endif

      if (((PVMMCB) hCurVM) -> CB_VM_Status & VMStat_Background)
      {
         PVDDCB pCurVMCB ;

         pCurVMCB = (PVDDCB)(hCurVM + dwVidCBOff) ;

         if ((0 == (pCurVMCB -> VDD_Flags & fVDD_WaitAtt)) &&
             (pCurVMCB -> VDD_Flags & fVDD_NoSaveRes))
         {
            VDD_VMMem_FreePages( hCurVM, pCurVMCB ) ;
            VDD_Proc_BGrndNotify( hCurVM, pCurVMCB ) ;
         }
      }
   }

} // VDD_Proc_NotifyBkgndVMs()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_ChkBkgndUsage
//  
//  Description:
//      This routine is called by VDD_VM_Mem_Set_Device_Focus and
//      VDD_VM_Mem_End_Mode_Chg for the current focus VM.  It is used
//      to check to see if the focus VM has all video pages
//      visible.  If all pages are visible, then any background
//      VM's which needs demand paged video memory must be
//      suspended.	If all pages are not visible, then any VMs
//      which the VDD might have previously suspended can be
//      resumed again.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Proc_ChkBkgndUsage
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if (VDD_PHMem_VMUsingAll( hVM, pVMCB ))
      VDD_Proc_SuspendBkgndVMs( hVM, pVMCB ) ;
   else
      VDD_Proc_ResumeBkgndVMs( hVM ) ;

} // VDD_Proc_ChkBkgndUsage()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_SuspendBkgndVMs
//  
//  Description:
//      Go thru VM list and suspend any background VM's which
//      need video memory (not text, CGA graphics or 256 color
//      graphics modes) and don't know how to restore themselves.
//
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM's control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Proc_SuspendBkgndVMs
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   DWORD  hCurVM ;

   if (pVMCB -> VDD_TFlags & fVT_NoSuspend)
      return ;

   for (hCurVM = vmmGetNextVMHandle( hVM );
        hCurVM != hVM;
        hCurVM = vmmGetNextVMHandle( hCurVM ))
   {
      // suspend any background VM's that need video memory

      if (((PVMMCB) hCurVM) -> CB_VM_Status & VMStat_Background)
      {
         PVDDCB pCurVMCB ;

         pCurVMCB = (PVDDCB) (hCurVM + dwVidCBOff) ;

         if (0 == (pCurVMCB -> VDD_Flags & fVDD_WaitAtt))
         {
            if (pCurVMCB -> VDD_Flags & fVDD_CanRestore)
            {
               if (pCurVMCB -> VDD_TFlags & fVT_Notify_at_PF)
                  VDD_Proc_BGrndNotify( hCurVM, pCurVMCB ) ;
               continue ;
            }
            else
            {
               if (!VDD_VMMem_CanUseSaveMem( hCurVM, pCurVMCB ))
               {
                  if (vmmSuspendVM( hCurVM ))
                  {
                     TRACEMSG( "suspended VM" ) ;

                     pCurVMCB -> VDD_Flags |= fVDD_WaitAtt ;
                     pCurVMCB -> VDD_EFlags |= fVDE_NoFGrnd ;
                     dwVidFlags |= fVid_VMsSuspended ;

                     if (0 == (pCurVMCB -> VDD_TFlags & fVT_SuprsMsg))
                        VDD_Error_NoFGrnd( hCurVM, pCurVMCB ) ;
                  }
               }

            }

         }

      }

   }

} // VDD_Proc_SuspendBkgndVMs()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_ResumeBkgndVMs
//  
//  Description:
//      Go thru VM list and resume any background VM's which
//      we previously suspended in VDD_Proc_Suspend_Bkgnd_VMs.
//      Also schedule a notification event for any VM's which
//      know how to restore their own screens and already got
//      a background INT 2F notification from us in VDD_PFault.
//  
//  Parameters:
//      DWORD hVM
//         handle to current VM
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Proc_ResumeBkgndVMs
(
    DWORD           hVM
)
{
   DWORD  hCurVM ;

   if (0 == (dwVidFlags & fVid_VMsSuspended))
      return ;

   dwVidFlags &= ~fVid_VMsSuspended ;

   hCurVM = hVM ;
   do
   {
      if (((PVMMCB) hCurVM) -> CB_VM_Status & VMStat_Background)
      {
         PVDDCB  pVMCB ;

         pVMCB = (PVDDCB) (hCurVM + dwVidCBOff) ;

         if (pVMCB -> VDD_Flags & fVDD_WaitAtt)
         {
            if (pVMCB -> VDD_Flags & fVDD_CanRestore)
            {
               if ((hCurVM == hVidFocusVM) ||
                   (0 == (pVMCB -> VDD_Flags & fVDD_NoSaveRes)))
               {
                  VDD_VMMem_DisableNull( hCurVM, pVMCB ) ;
                  pVMCB -> VDD_Flags &= ~fVDD_WaitAtt ;
                  if (0 == pVMCB -> VDD_NotifyEvent)
                     pVMCB -> VDD_NotifyEvent =
                        vmmScheduleVMEvent( hCurVM, pVDD_Proc_FGrndInt2F, 0 ) ;

               }
               else
                  dwVidFlags |= fVid_VMsSuspended ;
            }
            else
            {
               pVMCB -> VDD_Flags &= ~fVDD_WaitAtt ;
               vmmNoFailResumeVM( hCurVM ) ; 
            }
         }
      }
      hCurVM = vmmGetNextVMHandle( hCurVM ) ;
   }
   while (hCurVM != hVM) ;

} // VDD_Proc_ResumeBkgndVMs()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_ForceCurVMFgrnd
//  
//  Description:
//  
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Proc_ForceCurVMFgrnd
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if ((pVMCB -> VDD_Flags & fVDD_WaitAtt) &&
       (pVMCB -> VDD_Flags & fVDD_CanRestore))
   {
      VDD_VMMem_DisableNull( hVM, pVMCB ) ;
      pVMCB -> VDD_Flags &= ~fVDD_WaitAtt ;
      VDD_Proc_FGrndINT2F( hVM, 0 ) ;
   }

} // VDD_Proc_ForceCurVMFgrnd()

//---------------------------------------------------------------------------
//  End of File: vddproc.c
//---------------------------------------------------------------------------
