//---------------------------------------------------------------------------
//
//  Module: vddvga.c
//
//  Purpose:
//     VGA specific routines for VDD.
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              3/25/93   BryanW      Rewritten in 'C'.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vddvga.h"
#include "portio.h"

#pragma data_seg( "_LDATA", "CODE" ) 

//=============================================================================
//                             L C O D E
//=============================================================================

//--------------------------------------------------------------------------
//  
//  VOID VDD_VGA_SaveRegs
//  
//  Description:
//      Saves the state of the VGA by reading the registers.  Note that
//      the VDD_VGA_Savexxxx routines must be called in specific order.
//  
//  Parameters:
//      None.
//
//  Comments:
//     hVidMemCVM = VM current running in VGA
//  
//  Return (VOID):
//     VGA index registers restored, VDD_Stt structure modified
//
//  History:   Date       Author      Comment
//             10/21/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VGA_SaveRegs()
{
   PVDDCB  pVMCB ;
   WORD    wPort ;

   if (hVidMemCVM == hVidCRTCVM)
   {
      pVMCB = (PVDDCB) (hVidMemCVM + dwVidCBOff) ;

#ifdef DEBUG
      if (pVMCB -> VDD_Flags & fVDD_NeedRestore)
         DEBUGMSG( "Saving hardware regs before changed state restored\r\n" ) ;
#endif

      wPort = VDD_State_GetCRTCIndexPort() ;
      VDD_OEM_PreSaveRegs( pVMCB, wPort ) ;  
      VDD_VGA_SaveMisc( pVMCB, wPort ) ;      // save feat, misc, etc.
      VDD_VGA_SaveCRTC( pVMCB, wPort ) ;      // save CRTC
      VDD_VGA_SaveGrp( pVMCB ) ;              // save graphics controller
      VDD_VGA_SaveSeq( pVMCB ) ;              // save sequencer
      VDD_OEM_SaveRegs( pVMCB, wPort ) ;      // save OEM specific
      VDD_VGA_SaveAttr( pVMCB, wPort ) ;      // save attrib controller

      // Conditionally save the DAC

      if (dwVidFlags & fVid_DOff)
         VDD_VGA_SaveDAC( pVMCB ) ;           // save DAC
      VDD_OEM_PostSaveRegs( pVMCB, wPort ) ;

   }

} // VDD_VGA_SaveRegs()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VGA_SaveAttr
//  
//  Description:
//     Saves the state of the Attribute controller by reading the
//     registers.
//  
//  Parameters:
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//      WORD wPort
//         3B4 or 3D4
//
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/21/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VGA_SaveAttr
(
    PVDDCB          pVMCB,
    WORD            wPort
)
{
   BYTE   bReg ;
   UINT   i, iSize ;
   PBYTE  pStorage ;

   wPort += (pStatColr - pCRTCIndxColr) ;

   // Set toggle to index register

   inp( wPort ) ;

   // globally turned off or index saved and mod'd?
   // if so, don't read again

   if ((0 == (dwVidFlags & fVid_DOff)) &&
       (0 == (pVMCB -> VDD_Flags & fVDD_AxRd)))
   {
      // save attribute index register

      bReg = inp( pAttr ) ;
      pVMCB -> VDD_Stt.A_Indx &= fVAI_Indx ;
      bReg &= ~fVAI_Indx ;
      pVMCB -> VDD_Stt.A_Indx |= bReg ;
   }

   iSize =
      FIELDOFFSET( REGSTATE, A_Color ) - FIELDOFFSET( REGSTATE, A_Pal ) + 1 ;

   pStorage = (PBYTE) &pVMCB -> VDD_Stt.A_Pal ;

   // Turned off?  If so, read all the regs, otherwise
   // don't read the palette.

   bReg = 0 ;
   if (0 == (dwVidFlags & fVid_DOff))
   {
      // If screen's on, leave it on...
      if (pVMCB -> VDD_Stt.A_Indx & fVAI_ScOn)
         bReg |= fVAI_ScOn ;
      pStorage += 0x10 ;
      iSize -= 0x10 ;
      bReg |= 0x10 ;
   }

   // Wait for retrace (only when necessary)

   if ((0 == (dwVidFlags & fVid_DOff)) &&
       (0 == (bReg & 0x10)))
   {
      for (i = 0;
           ((inp( wPort ) & 0x08) == 0) && i < 0x10000;
           i++) ;
   }

   // Loop through and save regs...

   for (i = 0; i < iSize; i++)
   {
      // Set toggle to index register

      inp( wPort ) ;
      outp( pAttr, bReg++ ) ;
      *pStorage++ = inp( pAttr + 1 ) ;
   }

   // Set toggle to index register

   inp( wPort ) ;

   bReg = pVMCB -> VDD_Stt.A_Indx ;
   bReg &= 0xFF - fVAI_Indx ;

   // If video is off, keep it off..

   if (dwVidFlags & fVid_DOff)
      bReg &= ~fVAI_ScOn ;

   outp( pAttr, bReg ) ;

   // Reset toggle to index, if need be

   if (pVMCB -> VDD_Stt.A_Indx & fVAI_Indx)
      inp( wPort ) ;

} // VDD_VGA_SaveAttr()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VGA_SaveMisc
//  
//  Description:
//      Reads the Miscellaneous register and other OEM specific registers.
//
//  Parameters:
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//      WORD wPort
//         3B4 or 3D4
//
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/21/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VGA_SaveMisc
(
    PVDDCB          pVMCB,
    WORD            wPort
)
{

   wPort &= 0xFF00 ;
   wPort |= (pMiscRead & 0x00FF) ;

   pVMCB -> VDD_Stt.V_Misc = inp( wPort ) ;

   wPort &= 0xFF00 ;
   wPort |= (pFeatRead & 0x00FF) ;

   pVMCB -> VDD_Stt.V_Feat = inp( wPort ) ;

} // VDD_VGA_SaveMisc()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VGA_SaveCRTC
//  
//  Description:
//      Reads the CRTC registers.
//
//  Parameters:
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//      WORD wPort
//         3B4 or 3D4
//
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/21/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VGA_SaveCRTC
(
    PVDDCB          pVMCB,
    WORD            wPort
)
{

   UINT   i, iSize ;
   PBYTE  pStorage ;

   pVMCB -> VDD_Stt.C_Indx = inp( wPort ) & mx_CRTC ;

   pStorage = (PBYTE) &pVMCB -> VDD_Stt.crtc ;
   iSize =
      FIELDOFFSET( CRTC, C_LnCmp ) - FIELDOFFSET( CRTC, C_HTotal ) + 1 ;

   for (i = 0; i < iSize; i++)
   {
      // Select a register

      outp( wPort, i ) ;

      // Read and store it

      *pStorage++ = inp( wPort + 1 ) ;
   }

   outp( wPort, pVMCB -> VDD_Stt.C_Indx ) ;

} // VDD_VGA_SaveCRTC()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VGA_SaveGrp
//  
//  Description:
//      Reads the Graphics controller registers.
//
//  Parameters:
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/21/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VGA_SaveGrp
(
    PVDDCB          pVMCB
)
{
   UINT   i, iSize ;
   PBYTE  pStorage ;

   pVMCB -> VDD_Stt.G_Indx = inp( pGrpIndx ) & mx_Grp ;

   pStorage = (PBYTE) &pVMCB -> VDD_Stt.G_SetRst ;
   iSize =
      FIELDOFFSET( REGSTATE, G_Mask ) - FIELDOFFSET( REGSTATE, G_SetRst ) + 1 ;

   for (i = 0; i < iSize; i++)
   {
      // Select a register

      outp( pGrpIndx, i ) ;

      // Read and store it

      *pStorage++ = inp( pGrpIndx + 1 ) ;
   }

   outp( pGrpIndx, pVMCB -> VDD_Stt.G_Indx ) ;

} // VDD_VGA_SaveGrp()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VGA_SaveSeq
//  
//  Description:
//      Reads the Sequencer registers.
//
//  Parameters:
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/21/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VGA_SaveSeq
(
    PVDDCB          pVMCB
)
{
   UINT   i, iSize ;
   PBYTE  pStorage ;

   pVMCB -> VDD_Stt.S_Indx = inp( pSeqIndx ) & mx_Seq ;

   pStorage = (PBYTE) &pVMCB -> VDD_Stt.S_Rst ;
   iSize =
      FIELDOFFSET( REGSTATE, S_MMode ) - FIELDOFFSET( REGSTATE, S_Rst ) + 1 ;

   for (i = 0; i < iSize; i++)
   {
      // Select a register

      outp( pSeqIndx, i ) ;

      // Read and store it

      *pStorage++ = inp( pSeqIndx + 1 ) ;
   }

   outp( pSeqIndx, pVMCB -> VDD_Stt.S_Indx ) ;

} // VDD_VGA_SaveSeq()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VGA_SaveDAC
//  
//  Description:
//      Reads the Digital to Analog Converter (DAC) registers.
//
//  Parameters:
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/21/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VGA_SaveDAC
(
    PVDDCB          pVMCB
)
{
   UINT   i ;
   PBYTE  pStorage ;

   pStorage = pVMCB -> VDD_DAC ;

   for (i = 0; i < 256; i++)
   {
      // Set index then read data

      *pStorage++ = (BYTE) i ;
      outp( pDACRindx, i ) ;

      // Get red...

      *pStorage++ = inp( pDACData ) ;

      // Get green...

      *pStorage++ = inp( pDACData ) ;

      // Get blue...

      *pStorage++ = inp( pDACData ) ;
   }

} // VDD_VGA_SaveDAC()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VGA_RestoreDAC
//  
//  Description:
//      Restores the DAC.
//
//  Parameters:
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/21/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VGA_RestoreDAC
(
    PVDDCB          pVMCB
)
{
   UINT   i, iSize ;
   PBYTE  pStorage ;

   pStorage = pVMCB -> VDD_DAC ;

   iSize = DAC_State_Size / 4 ;

   // If message mode, only restore 16 colors

   if (dwVidFlags & fVid_MsgA)
      iSize = DAC_Msg_State_Size / 4 ;

   for (i = 0; i < iSize; i++)
   {
      // Set index then read data

      outp( pDACWindx, *pStorage++ ) ;

      // Output red...

      outp( pDACData, *pStorage++ ) ;

      // Output green...

      outp( pDACData, *pStorage++ ) ;

      // Output blue...

      outp( pDACData, *pStorage++ ) ;
   }

} // VDD_VGA_RestoreDAC()

//---------------------------------------------------------------------------
//  End of File: vddvga.c
//---------------------------------------------------------------------------


