//---------------------------------------------------------------------------
//
//  Module: vddvmmem.c
//
//  Purpose:
//     This module handles all of the manipulation of a VM's video memory
//     space. This includes mapping of physical device, mapping normal RAM
//     for emulation of physical device and null pages.  The pages are
//     disabled and hooked when they have not been allocated to the VM
//     and when the mode of the controller has changed.
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              4/11/93   BryanW      Rewritten in 'C'.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vddvga.h"
#include "vddvmmem.h"
#include "vddphmem.h"
#include "portio.h"

#pragma data_seg( "_LDATA", "CODE" ) 

//=============================================================================
//                    G L O B A L   D A T A   A R E A 
//=============================================================================

DWORD      dwReservedROMPage ;
DWORD      dwVidPageAllocFlags = PageZeroInit | PageLocked ;
DWORD      dwVidUnaccessedPage = 0 ;

EVENTPROC  pVDD_Error_CvtFullScrn ;

//=============================================================================
//                             L C O D E
//=============================================================================

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_AllocPageMaps
//
//  Description:
//     Allocate the 2 pagemap arrays that record which page handles
//     are allocated for each virtual page # and flags for each
//     virtual page #.
//
//  Parameters:
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     DWORD dwVidPages
//        new VDD_VideoPages value
//
//  History:   Date       Author      Comment
//              4/11/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_AllocPageMaps
(
    PVDDCB          pVMCB,
    DWORD           dwVidPages
)
{
   DWORD  dwTemp ;

   pVMCB -> VDD_VideoPages = dwVidPages ;
   if (NULL ==
       (dwTemp = vmmHeapAllocate( dwVidPages << 2, HeapZeroInit )))
   {
      DEBUGMSG( "VDD ERROR: Could not allocate pagemap\r\n" ) ;
      return ( FALSE ) ;
   }
   pVMCB -> VDD_PageMap = (PDWORD) dwTemp ;

   if (NULL ==
       (dwTemp = vmmHeapAllocate( dwVidPages, HeapZeroInit )))
   {
      DEBUGMSG( "VDD ERROR: Could not allocate pageflags\r\n" ) ;
      return ( FALSE ) ;
   }
   pVMCB -> VDD_PageFlags = (PBYTE) dwTemp ;

   return ( TRUE ) ;

} // end of VDD_VMMem_AllocPageMaps()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_GrowPageMaps
//
//  Description:
//     Attempts to grow the 2 pagemap arrays for the given VM.
//
//  Parameters:
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     DWORD dwVidPages
//        new VDD_VideoPages value - 1
//
//  Return Value:
//     BOOL
//        TRUE if successful, FALSE otherwise
//
//  History:   Date       Author      Comment
//              4/13/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_GrowPageMaps
(
    PVDDCB          pVMCB,
    DWORD           dwVidPages
)
{
   DWORD dwTemp ;

   if (NULL == (dwTemp = vmmHeapReAllocate( (DWORD) pVMCB -> VDD_PageMap,
                                            (dwVidPages + 1) << 2,
                                            HeapZeroInit )))
   {
      DEBUGMSG( "VDD ERROR: Could not allocate pagemap\r\n" ) ;
      return ( FALSE ) ;
   }
   pVMCB -> VDD_PageMap = (PDWORD) dwTemp ;
   if (NULL == (dwTemp = vmmHeapReAllocate( (DWORD) pVMCB -> VDD_PageFlags,
                                            dwVidPages + 1,
                                            HeapZeroInit )))
   {
      DEBUGMSG( "VDD ERROR: Could not allocate pageflags\r\n" ) ;
      return ( FALSE ) ;
   }
   pVMCB -> VDD_PageFlags = (PBYTE) dwTemp ;
   pVMCB -> VDD_VideoPages = dwVidPages + 1 ;
   return ( TRUE ) ;

} // end of VDD_VMMem_GrowPageMaps()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_VMCriticalInit
//
//  Description:
//     Map physical pages for video BIOS pages 6 & 7 into the
//     current VM, if the VDD reserved the pages in the
//     VDD_VMMem_DeviceInit (usually these will be pages C6 & C7
//     or E6 & E7).
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//  Return Value:
//     None.
//
//  History:   Date       Author      Comment
//              4/13/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_VMCriticalInit
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if (pVMCB -> VDD_TFlags & fVT_RsrvC6C7)
   {
      // Map Physical == Linear for BIOS pages

      vmmPhysIntoV86( dwReservedROMPage, hVM, dwReservedROMPage, 2, 0 ) ;
   }

   if (pVMCB -> VDD_TFlags & fVT_ResvB0)
   {
      // Map Physical == Linear from B0 to B7

      vmmPhysIntoV86( 0xB0, hVM, 0xB0, 8, 0 ) ;
   }

} // end of VDD_VMMem_VMCriticalInit()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_AllocVisiblePage
//
//  Description:
//     This routine is called in by VDD_VM_Mem_Set_Device_Focus
//     to allocate a page of visible video memory, or a font page.
//     It handles unmapping save memory pages that might be mapped
//     into the VM, if it is in a text or CGA mode.  It also
//     handles moving pages around, if the page indentified in
//     EAX is already mapped to a different physical page.  After
//     a page is allocated, its saved contents are restored by
//     a call to VDD_VM_Mem_Restore_Page.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     UINT uPageID
//        linear page ID
//
//  History:   Date       Author      Comment
//              4/24/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_AllocVisiblePage
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uPageID
)
{
   DWORD  hOwnerVM, hPage, hNewPage ;

   if (pVMCB -> VDD_PageFlags[ uPageID ] & fPageSaveMem)
      VDD_VMMem_UnmapSaveMem( hVM, pVMCB, uPageID ) ;

   if (VDD_PHMem_GetPageOwner( hVM, pVMCB, uPageID, &hOwnerVM, &hPage ))
   {
      // Owned.. check that VM owns page

      if (hVM == hOwnerVM)
      {
         // If page id is the same, we're done.

         if (((PPAGEDATA) hPage) -> PDS_virtual_id == uPageID)
         {
            VDD_PHMem_UpdateLRU( pVMCB, hPage ) ;
            return ;
         }
         else
         {
            // Check to see if page id and virtual id share same
            // physical memory, if so allocate a new virtual page.
            // Otherwise if it was lower, we should have already
            // caused it to move, if it's higher we just free it
            // rather than mess with further conflicts.

            if (VDD_PHMem_SharedPage( hVM, pVMCB, uPageID,
                                      ((PPAGEDATA) hPage) -> PDS_virtual_id ))
               goto AVP_AllocNew ;
         }
      }
      if (!VDD_PHMem_FreePage( hPage ))
      {
         DEBUGMSG( "failed to free visible page for new focus\r\n" ) ;
         VDD_Error_NoPagesAvail() ;
         return ;
      }
   }

   if (NULL != (hNewPage = ((PDWORD) pVMCB -> VDD_PageMap)[ uPageID ]))
   {
      if (NULL ==
            (hNewPage = VDD_PHMem_ChangePage( hVM, pVMCB, uPageID,
                                              hNewPage, NULL )))
      {
         DEBUGMSG( "Failed to move a video page\r\n" ) ;
         VDD_Error_NoPagesAvail() ;
         return ;
      }
   }
   else
   {
AVP_AllocNew:
      if (NULL ==
            (hNewPage =
               VDD_PHMem_AllocVideoPage( hVM, pVMCB, uPageID,
                                         VDD_VMMem_PageHandler )))
      {
         DEBUGMSG( "VDD_VMMem_AllocVisiblePage failed to alloc page!\r\n" ) ;
         VDD_Error_NoPagesAvail() ;
         return ;
      }
   }

   pVMCB -> VDD_PageMap[ uPageID ] = hNewPage ;
   pVMCB -> VDD_PageFlags[ uPageID ] &= fPageGblFlags ;
   VDD_VMMem_RestorePage( hVM, pVMCB, hNewPage,
                          &pVMCB -> VDD_PageFlags[ uPageID ] ) ;

   VDD_PHMem_UpdateLRU( pVMCB, hNewPage ) ;

   return ;

} // end of VDD_VMMem_AllocVisiblePage()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_ChkVideoPages
//
//  Description:
//     Grow page maps, if needed based on call to 
//     VDD_PHMem_GetVisiblePages.
//
//  Parameters:
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     UINT uFirstPage
//        first page ID
//
//     UINT uPages
//        count of pages
//
//  Return Value:
//     TRUE if page maps large enough, 
//     FALSE if page maps too small and realloc failed
//
//  History:   Date       Author      Comment
//              4/24/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_ChkVideoPages
(
    PVDDCB          pVMCB,
    UINT            uFirstPage,
    UINT            uPages
)
{
   if (pVMCB -> VDD_VideoPages > uFirstPage + uPages)
      return ( TRUE ) ;

   TRACEPMSG( "Growing page maps in VDD_VMMem_ChkVideoPages (#EAX)\r\n",
              uFirstPage + uPages ) ;

   return ( VDD_VMMem_GrowPageMaps( pVMCB, uFirstPage + uPages - 1 ) ) ;

} // end of VDD_VMMem_ChkVideoPages()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_SetDeviceFocus
//
//  Description:
//     Allocate all visible video pages for the new focus VM and
//     restore the VM's saved video memory into these allocated
//     pages.  If the VM already owns video pages for some of
//     these visible pages, then they will be used or moved as
//     necessary to force the visible pages to all be contiguous
//     video pages.  If the new focus has all availble video pages
//     visible, then we must suspend all background VM's which
//     need video memory.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//  History:   Date       Author      Comment
//              5/10/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_SetDeviceFocus
(
    DWORD           hVM
)
{
   BOOL    fGraphics ;
   BYTE    bFontPages ;
   UINT    i, uVisFirst, uVisPages ;
   PVDDCB  pVMCB ;

   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;

   TRACEMSG( "A" ) ;

   // VDD_VMMem_StateVMDebug() ;

   VDD_Proc_NotifyBkgndVMs( hVM, pVMCB ) ;
   VDD_VMMem_MemC_AboutToChange( hVM, pVMCB ) ;
   VDD_State_BeginSRSection() ;

#ifdef SysVMin2ndBank
   if ((pVMCB -> VDD_TFlags & fVT_SysVMnot1stBank) &&
         vmmTestSysVMHandle( hVM ))
   {
      vmmPhysIntoV86( 0xA0, hVM, 0xA0, 16, 0 ) ;
      goto SDF_Exit ;
   }
#endif

   fGraphics =
      VDD_PHMem_GetVisiblePages( hVM, pVMCB, &uVisFirst, &uVisPages ) ;

   if (uVisPages > 16)
   {
      uVisPages = 16 ;
      DEBUGMSG( "visible pages greater than 16\r\n" ) ;
   }

   if (!VDD_VMMem_ChkVideoPages( pVMCB, uVisFirst, uVisPages ))
   {
      // VDD_Error_NoMainMem() ;

      // Adjust visible down by alloc'ed and
      // adjust for first page.

      uVisPages = pVMCB -> VDD_VideoPages ;
      uVisPages -= uVisFirst ;
   }

   pVMCB -> VDD_first_page = (BYTE) uVisFirst ;
   pVMCB -> VDD_visible_pages = (BYTE) uVisPages ;

   // Allocate visible pages...

   for (i = 0; i < uVisPages; i++)
      VDD_VMMem_AllocVisiblePage( hVM, pVMCB, uVisFirst + i ) ;

   if (!fGraphics)
   {
      bFontPages = pVMCB -> VDD_fontpages ;
      for (i = 0; bFontPages != 0; i += 2)
      {
         if (bFontPages & 0x01)
         {
            VDD_VMMem_AllocVisiblePage( hVM, pVMCB, i ) ;
            VDD_VMMem_AllocVisiblePage( hVM, pVMCB, i + 1 ) ;
         }
         bFontPages >>= 1 ;
      }
   }

#ifdef SysVMin2ndBank
SDF_Exit:
#endif
   VDD_PHMem_LRUSweep() ;
   VDD_PHMem_SetLatchAddr() ;
   VDD_Proc_ChkBkgndUsage( hVM, pVMCB ) ;
   VDD_State_EndSRSection() ;

   if (vmmTestCurVMHandle( hVM ))
   {
      VDD_State_SetMemCOwner( hVM, pVMCB ) ;
      VDD_Proc_ForceCurVMFgrnd( hVM, pVMCB ) ;
   }

} // end of VDD_VMMem_SetDeviceFocus()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_CreateVM
//
//  Description:
//     Unmap all video memory in the VM's page table and
//     allocate initial video save memory for the VM.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//  Return Value:
//     BOOL
//        TRUE if successful, FALSE otherwise.
//
//  History:   Date       Author      Comment
//              4/13/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_CreateVM
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   int    i ;
   DWORD  dwMemAlloc ;

   pVMCB -> VDD_Scroll_Cntr = LOWORD( dwVidScrollFreq ) ;
   if (!vmmTestSysVMHandle( hVM ))
      if (!VDD_VMMem_AllocPageMaps( pVMCB, dwVidPages ))
      {
         DEBUGMSG( "Failed to allocate pagemap tables for new VM\r\n" ) ;
         return ( FALSE );
      }

   vmmModifyPageBits( hVM, 0xA0, 16, ~P_AVAIL, NULL, PG_HOOKED, NULL ) ;
   vmmModifyPageBits( hVM, 0xB8, 8, ~P_AVAIL, NULL, PG_HOOKED, NULL ) ;

   if (pVMCB -> VDD_TFlags & fVT_Mono)
      vmmModifyPageBits( hVM, 0xB0, 8, ~P_AVAIL, NULL, PG_HOOKED, NULL ) ;

   dwMemAlloc = 0xA0A0A0A ;   // default to 4 planes of 10 pages

   if (0 == pVMCB -> VDD_PIF & fVidHghRsGrfxMd)
   {
      dwMemAlloc = 4 ;
      if (0 == pVMCB -> VDD_PIF & fVidLowRsGrfxMd)
      {
         dwMemAlloc = bVidInitialTextRows * 32 * 5 + P_SIZE - 1 ;

         dwMemAlloc >>= 12 ;           // number of pages
         dwMemAlloc |= 0x20000 ;       // assume text mode alloc
      }
   }

   for (i = 0; i < 4; i++)
   {
      pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_handle =
         NULL ;
      pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_addr =
         NULL ;
      pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_size =
         (dwMemAlloc >> (8 * i)) & 0xFF ;
      pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_alloced =
         NULL ;
   }

   if (!VDD_VMMem_ReAlloc( hVM, pVMCB, &pVMCB -> VDD_MStt ))
   {
      DEBUGMSG( "Failed to allocate save memory for new VM\r\n" ) ;
      VDD_VMMem_DMain( pVMCB ) ;
      return ( FALSE ) ;
   }
   else
   {
      pVMCB -> VDD_fontpages = 1 ;
      pVMCB -> VDD_Flags = fVDD_MInit ;
   }
   return ( TRUE ) ;
   
} // end of VDD_VMMem_CreateVM()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_DestroyVM
//
//  Description:
//     Deallocate any video pages which the VM has allocated and
//     deallocate the video save memory.
//
//  Parameters:
//     DWORD hVM
//        handle to VM being destroyed.
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_DestroyVM
(
    DWORD           hVM
)
{
   PVDDCB  pVMCB ;

   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;

   // Release all the memory now...

   pVMCB -> VDD_PIF &= ~fVidRetainAllo ;

   VDD_VMMem_FreePages( hVM, pVMCB ) ;
   VDD_VMMem_DMain( pVMCB ) ;

   VDD_Grab_DeleteTextCopyBuf( hVM, pVMCB ) ;

} // end of VDD_VMMem_DestroyVM()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_VMResume
//
//  Description:
//     Lock the VM's video save memory
//
//  Parameters:
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//  Return Value:
//     TRUE if successful, otherwise FALSE if can't resume VM
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_VMResume
(
    PVDDCB          pVMCB
)
{
   int    i ;
   DWORD  hMem, dwSize ;

   for (i = 0; i < 4; i++)
   {
      hMem = pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_handle ;
      dwSize = pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_alloced ;
      if (hMem)
      {
         if (!vmmPageLock( hMem, dwSize, 0, PageLockedIfDP ))
         {
            VDD_VMMem_UnlockSaveMem( pVMCB, i ) ;
            return ( FALSE ) ;
         }
      }
   }
   return ( TRUE ) ;

} // end of VDD_VMMem_VMResume()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_VMSuspend
//
//  Description:
//     Unlock the VM's video save memory
//
//  Parameters:
//     PVDDCB pVMCB
//        pointer to VM control block
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_VMSuspend
(
    PVDDCB          pVMCB
)
{
   VDD_VMMem_UnlockSaveMem( pVMCB, 4 ) ;

} // end of VDD_VMMem_VMSuspend()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_UnlockSaveMem
//
//  Description:
//     Unlocks the save memory planes starting with the first plane.
//
//  Parameters:
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     UINT uPlanes
//        number of planes to unlock
//
//  History:   Date       Author      Comment
//              4/22/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_UnlockSaveMem
(
    PVDDCB          pVMCB,
    UINT            uPlanes
)
{
   int    i ;
   DWORD  hMem, dwSize ;

   for (i = 0; i < uPlanes; i++)
   {
      hMem = pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_handle ;

      if (hMem)
      {
         dwSize = pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_alloced ;
         vmmPageUnLock( hMem, dwSize, 0, PageLockedIfDP ) ;
      }
   }

} // end of VDD_VMMem_UnlockSaveMem()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_DeAlloc
//
//  Description:
//     Deallocate all video save memory pages allocated to a
//     MEMSTATE structure (either VDD_MStt or VDD_MSttCopy)
//
//  Parameters:
//     PMEMSTATE pMS
//        pointer to MEMSTATE structure to reallocate
//
//  History:   Date       Author      Comment
//              4/22/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_DeAlloc
(
    PMEMSTATE       pMS
)
{
   int  i ;

   for (i = 0; i < 4; i++)
   {
      pMS -> MSS_planes[ i ].MSS_plane_addr = NULL ;
      pMS -> MSS_planes[ i ].MSS_plane_size = 0 ;
      pMS -> MSS_planes[ i ].MSS_plane_alloced = 0 ;
      vmmPageFree( pMS -> MSS_planes[ i ].MSS_plane_handle, 0 ) ;
      pMS -> MSS_planes[ i ].MSS_plane_handle = NULL ;
   }

} // end of VDD_VMMem_DeAlloc()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_DMain
//
//  Description:
//     Deallocates all video save memory pages allocated to a
//     VDD_MStt.
//
//  Parameters:
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//  History:   Date       Author      Comment
//              4/22/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_DMain
(
    PVDDCB          pVMCB
)
{
   VDD_VMMem_DeAlloc( &pVMCB -> VDD_MStt ) ;

} // end of VDD_VMMem_DMain()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_SetPlaneSizes
//
//  Description:
//     Sets the new plane sizes.
//
//  Parameters:
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     PMEMSTATE pMS
//        pointer to MEMSTATE structure
//
//     WORD acpPlanes[]
//        array of count of pages for each plane
//
//  History:   Date       Author      Comment
//              4/24/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_SetPlaneSizes
(
    PVDDCB              pVMCB,
    PMEMSTATE           pMS,
    WORD                acpPlanes[]
)
{
   int    i ;
   PWORD  pwPS ;

   for (i = 0; i < 4; i++)
   {
      
      pwPS = &pMS -> MSS_planes[ i ].MSS_plane_size ;

      // Modify MSS_plane_size IFF:
      //
      //    1) PIF requested no mem shrink _AND_ new size > old size
      //
      //    OR
      //
      //    2) Saving RAM during mode change _OR_ new size > old size
      //

      if (((pVMCB -> VDD_PIF & fVidRetainAllo) && (acpPlanes[ i ] > *pwPS)) ||
         ((pVMCB -> VDD_Flags & fVDD_MdSvRAM) || (acpPlanes[ i ] > *pwPS)))
         *pwPS = acpPlanes[ i ] ;
   }

} // end of VDD_VMMem_SetPlaneSizes()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_AdjustPlaneSizes
//
//  Description:
//     Adjust the plane sizes needed to save the current video
//     memory state.  This routine just modifies the
//     MSS_plane_size fields in the Mem_State_Struc.  It does
//     not allocate or deallocate memory.	A subsequent call
//     to VDD_VM_Mem_Realloc can be made to change the actual
//     memory allocations.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     PMEMSTATE pMS
//        pointer to MEMSTATE structure
//
//     BOOL fCopy
//        copy memory?
//           FALSE if main memory alloc, TRUE if copy memory alloc
//
//  History:   Date       Author      Comment
//              4/24/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_AdjustPlaneSizes
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PMEMSTATE       pMS,
    BOOL            fCopy
)
{
   BYTE  bGraphicsMode ;
   UINT  uVisFirst, uVisPages ;
   WORD  acpPlanes[ 4 ] ;

   // Don't care about return value (graphics mode)

   VDD_PHMem_GetVisiblePages( hVM, pVMCB, &uVisFirst, &uVisPages ) ;

   // Compute # of pages for plane 0 & assume 0 for planes 1 & 2

   acpPlanes[ 0 ] = uVisFirst + uVisPages ;
   acpPlanes[ 1 ] = acpPlanes[ 2 ] = 0 ;

   if (pVMCB -> VDD_Stt.S_MMode & fSeq4Chain4)
   {
      // We're in chain 4 mode (256 color)

      acpPlanes[ 3 ] = 0 ;
      VDD_VMMem_SetPlaneSizes( pVMCB, pMS, acpPlanes ) ;
      return ;
   }

   bGraphicsMode = pVMCB -> VDD_Stt.G_Misc ;

   // If CGA compatibility enabled, mode 4, 5 or 6 use 1 plane

   if (pVMCB -> VDD_Stt.crtc.C_Mode & 1)
   {
      bGraphicsMode &= fGrp6Char | fGrp6Chain ;

      if (bGraphicsMode == (fGrp6Char | fGrp6Chain))
      {
         // chained odd to even graphics mode

         acpPlanes[ 2 ] = acpPlanes[ 0 ] ;
         VDD_VMMem_SetPlaneSizes( pVMCB, pMS, acpPlanes ) ;
         return ;
      }

      // Check for odd/even mode (text & CGA)

      if (pVMCB -> VDD_Stt.S_MMode & fSeq4SqAd)
      {
         // planes 1, 2 & 3 need all pages

         acpPlanes[ 1 ] = acpPlanes[ 2 ] = acpPlanes[ 3 ] = acpPlanes[ 0 ] ;
         VDD_VMMem_SetPlaneSizes( pVMCB, pMS, acpPlanes ) ;
         return ;
      }
   }

   // If graphics mode or copy memory then no font pages

   if ((bGraphicsMode & fGrp6Char) || fCopy)
      acpPlanes[ 3 ] = 0 ;
   else
   {
      BYTE bFontPages ;

      // determine number of font pages needed

      bFontPages = pVMCB -> VDD_fontpages ;

      // count fonts

      for (acpPlanes[ 2 ] = 0; bFontPages != 0;
            bFontPages >>= 1, acpPlanes[ 2 ]++) ;

      // convert to pages

      acpPlanes[ 2 ] <<= 1 ;
      acpPlanes[ 3 ] = 0 ;
   }
   VDD_VMMem_SetPlaneSizes( pVMCB, pMS, acpPlanes ) ;
   return ;

} // end of VDD_VMMem_AdjustPlaneSizes()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_ReAllocPlane
//
//  Description:
//     Change the memory allocations associated with a plane
//     in the Mem_State_Struc.  This routine handles deallocating
//     a plane's memory, if the size shrinks to 0.  And doing a
//     new allocation, if the plane currently doesn't own any
//     memory.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PMSSPLANEINFO pPI
//        pointer to MSSPLANEINFO structure
//
//  Return Value:
//     TRUE if successful, FALSE otherwise
//
//  History:   Date       Author      Comment
//              4/22/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_ReAllocPlane
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI
)
{
   DWORD  hMem, dwLinAddr ;

   hMem = NULL ;
   dwLinAddr = NULL ;

   if (pMPI -> MSS_plane_size)
   {
      if (pMPI -> MSS_plane_handle)
         dwLinAddr = vmmPageReAllocate( &hMem, pMPI -> MSS_plane_handle,
                                        pMPI -> MSS_plane_size,
                                        dwVidPageAllocFlags ) ;
      else
         dwLinAddr = vmmPageAllocate( &hMem, pMPI -> MSS_plane_size,
                                      PG_HOOKED, hVM, 0, 0, 0, 0,
                                      dwVidPageAllocFlags ) ;

      if (!hMem)
         return ( FALSE ) ;
   }
   else if (pMPI -> MSS_plane_handle)
      vmmPageFree( pMPI -> MSS_plane_handle, 0 ) ;

   pMPI -> MSS_plane_handle = hMem ;
   pMPI -> MSS_plane_addr = dwLinAddr ;
   pMPI -> MSS_plane_alloced = pMPI -> MSS_plane_size ;
   return ( TRUE ) ;

} // end of VDD_VMMem_ReAllocPlane()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_ReAlloc
//
//  Description:
//     Change the memory allocations associated with each plane
//     in the Mem_State_Struc.  This routine handles deallocating
//     a plane's memory if the size shrinks to 0.  A new allocation
//     is performed if the plane currently doesn't own any memory.
//
//  Parameters:
//     DWORD hVM
//        handle to current VM
//
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//     PMEMSTATE pMS
//        pointer to MEMSTATE structure to reallocate
//
//  Return Value:
//     TRUE if successful, FALSE otherwise
//
//  History:   Date       Author      Comment
//              4/22/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_ReAlloc
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PMEMSTATE       pMS
)
{
   int    i ;
   DWORD  dwSize, dwAlloced ;

   if (pVMCB -> VDD_TFlags & fVT_NoSaveResSysVM)
   {
      // Don't alloc mem for Sys VM

      if (vmmTestSysVMHandle( hVM ))
         return ( TRUE ) ;
   }
   pVMCB -> VDD_EFlags &= ~fVDE_NoMain ;

   for (i = 0; i < 4; i++)
   {
      dwSize = pMS -> MSS_planes[ i ].MSS_plane_size ;
      dwAlloced = pMS -> MSS_planes[ i ].MSS_plane_alloced ;

      // Check:
      // 1) If required == allocated, if so, no work to do.
      // 2) If allocated > required and we don't want to retain
      //    allocated memory, it needs to be shrunk.

      if ((dwSize != dwAlloced) &&
          ((dwAlloced > dwSize) && !(pVMCB -> VDD_PIF & fVidRetainAllo)))
      {
         if (pVMCB -> VDD_savemem_maps)
         {
            TRACEMSG( "VMMem_ReAlloc: calling disable to unmap save mem\r\n" ) ;
            VDD_VMMem_Disable( hVM, pVMCB ) ;
         }
         if (!VDD_VMMem_ReAllocPlane( hVM, &pMS -> MSS_planes[ i ] ))
         {
            TRACEMSG( "VMMem_ReAlloc: Failed to allocate memory for plane!\r\n" ) ;
            pVMCB -> VDD_EFlags |= fVDE_NoMain ;
            return ( FALSE ) ;
         }

         // If plane 0 _and_ we're in text mode _and_ the allocation
         // grew, init the new pages...

         if ((i == 0) && !(pVMCB -> VDD_Stt.G_Misc & fGrp6Char) &&
             (dwAlloced < pMS -> MSS_planes[ i ].MSS_plane_alloced))
         {
            int     j, nClear ;
            PDWORD  pRegion ;

            // Compute the number of dwords in the new region

            nClear = pMS -> MSS_planes[ i ].MSS_plane_alloced - dwAlloced ;

            // Convert from pages to dwords

            nClear <<= 12 - 2 ;

            // Compute offset into plane

            pRegion =
               (PDWORD) (pMS -> MSS_planes[ i ].MSS_plane_addr +
                         (dwAlloced << 12)) ;

            for (j = 0; j < nClear; j++)
               *pRegion++ = 0x07200720 ;
         }
      }
   }
   return ( TRUE ) ;

} // end of VDD_VMMem_ReAlloc()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_FontPage
//
//  Description:
//     This routine is only called by VDD_PHMem_UpdateLRU to
//     determine if a video page is a necessary part of the
//     visible screen of the focus VM.  It first checks for
//     visible pages and if the page being checked isn't
//     visible it calls here to see if it might be part of a
//     font.  An example is that the 2nd page is not visible
//     in 80 by 25 text, but it is part of the 1st font, so
//     it is required.
//
//  Parameters:
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     UINT uPageID
//        page id
//
//  Return Value:
//     Returns TRUE, if the page is part of a loaded text
//     mode font, otherwise FALSE.
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_FontPage
(
    PVDDCB              pVMCB,
    UINT                uPageID
)
{
   if (pVMCB -> VDD_Stt.G_Misc & fGrp6Char)
      return ( FALSE ) ;
   else
      return ( (pVMCB -> VDD_fontpages & (1 << (uPageID >> 1))) ? TRUE : FALSE ) ;

} // end of VDD_VMMem_FontPage()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_EndModeChg
//
//  Description:
//     Adjust memory allocation after mode change.
//
//  Parameters:
//
//  History:   Date       Author      Comment
//              5/ 2/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_EndModeChg
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   BYTE   bFontPages ;
   BOOL   fFreePage, fVisible ;
   DWORD  hPage ;
   UINT   i, uLinPage, uVisFirst, uVisPages ;

#ifdef DEBUG
   // TRACEMSG( "A" ) ;
   // VDD_StateVM_Debug() ;
#endif
   if (0 == pVMCB -> VDD_Flags & fVDD_NoSaveRes)
   {
      VDD_VMMem_AdjustPlaneSizes( hVM, pVMCB, &pVMCB -> VDD_MStt, FALSE ) ;
      VDD_VMMem_ReAlloc( hVM, pVMCB, &pVMCB -> VDD_MStt ) ;
   }
   if (hVM == hVidFocusVM)
   {
      VDD_PHMem_SetLatchAddr() ;
      VDD_PHMem_ChkBkgndUsage() ;
   }

   bFontPages = 0 ;
   if (!VDD_PHMem_GetVisiblePages( hVM, pVMCB, &uVisFirst, &uVisPages ))
      bFontPages = pVMCB -> VDD_fontpages ;

   if (!VDD_VMMem_ChkVideoPages( pVMCB, uVisFirst, uVisPages ))
   {
      uVisPages = pVMCB -> VDD_VideoPages ;
      uVisPages -= uVisFirst ;
   }

   for (i = 0; i < pVMCB -> VDD_VideoPages; i++)
   {
      // If allocated page is not visible, then free it

      fVisible = fFreePage = FALSE ;
      if (bFontPages & (1 << (i >> 1)))
         fVisible = TRUE ;
      else
         fVisible = ((i < uVisFirst) || (i >= uVisFirst + uVisPages)) ;

      fFreePage = !fVisible ;
      if (fVisible)
      {
         if (hVidCRTCVM != hVM)
            fFreePage = VDD_VMMem_VMCanUseSaveMem( pVMCB ) ;
      }

      if (fFreePage)
      {
         if (hPage = pVMCB -> VDD_PageMap[ i ])
         {
            if (!fVisible)
            {
               pVMCB -> VDD_PageFlags[ i ] &= ~fPageSaved ;
               uLinPage = ((PPAGEDATA) hPage) -> PDS_linear_page ;
               ((PPAGEDATA) hPage) -> PDS_linear_page = 0 ;
               if (uLinPage)
                  vmmModifyPageBits( hVM, uLinPage, 1, ~P_AVAIL, 0,
                                     PG_HOOKED, 0 ) ;
               ((PPAGEDATA) hPage) -> PDS_flags &= ~fPDS_Dirty ;
            }
            VDD_State_BeginSRSection() ;
            VDD_PHMem_FreePage( hPage ) ;
         }
      }
   }

   VDD_VMMem_DisableNull( hVM, pVMCB ) ;
   VDD_Grab_ClearModState( hVM, pVMCB ) ;
   VDD_State_EndSRSection() ;

} // end of VDD_VMMem_EndModeChg()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_ChangeSaveMem
//
//  Description:
//     Deallocate VM's save memory, if the VM has just enabled
//     the CanRestore bit and auto restore is disabled for the
//     VM (sys VM/other VMs).  Or, reallocate VM's save memory,
//     if the VM has just disabled the CanRestore bit and auto
//     restore is disabled for the VM.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_ChangeSaveMem
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if (pVMCB -> VDD_Flags & fVDD_NoSaveRes)
   {
      VDD_VMMem_DMain( pVMCB ) ;
   }
   else
   {
      VDD_VMMem_AdjustPlaneSizes( hVM, pVMCB, &pVMCB -> VDD_MStt, FALSE ) ;
      VDD_VMMem_ReAlloc( hVM, pVMCB, &pVMCB -> VDD_MStt ) ;
   }

} // end of VDD_VMMem_ChangeSaveMem()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_AdjustVisiblePages
//
//  Description:
//     Check and adjust visible pages allocation if necessary
//     based upon CRTC start address.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB 
//        pointer to VM control block.
//
//  History:   Date       Author      Comment
//              4/27/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_AdjustVisiblePages
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   int   i ;
   UINT  uVisFirst, uVisPages ;

   TRACEMSG( "Adjusting visible pages\r\n" ) ;

   VDD_PHMem_GetVisiblePages( hVM, pVMCB, &uVisFirst, &uVisPages ) ;

   pVMCB -> VDD_first_page = uVisFirst ;
   pVMCB -> VDD_visible_pages = uVisPages ;

#ifdef DEBUG
   if (uVisFirst + uVisPages > pVMCB -> VDD_VideoPages)
      DEBUGMSG( "Too many pages visible in VDD_VMMem_AdjustVisiblePages\r\n" ) ;
#endif

   if (uVisFirst + uVisPages > pVMCB -> VDD_VideoPages)
      uVisPages = pVMCB -> VDD_VideoPages - uVisFirst ;

   for (i = uVisFirst; i < uVisPages; i++)
      VDD_VMMem_AllocVisiblePage( hVM, pVMCB, i ) ;

   VDD_VMMem_ChkAlloc( hVM, pVMCB, i - 1 ) ;
   VDD_PHMem_SetLatchAddr() ;

} // end of VDD_VMMem_AdjustVisiblePages()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_CA_CheckPlane0
//
//  Description:
//     Checks plane 0 and reallocates if necessary.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//     UINT uAllocNeeded
//        new size for plane
//
//  Return Value:
//     TRUE if successful, FALSE otherwise
//
//  History:   Date       Author      Comment
//              4/29/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_CA_CheckPlane0
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uAllocNeeded
)
{
   // Check allocation of plane 0

   if (pVMCB -> VDD_MStt.MSS_planes[ 0 ].MSS_plane_size >= uAllocNeeded)
      return ( TRUE ) ;
   else
   {
      TRACEMSG( "Calling VDD_VMMem_ReAlloc from VDD_VMMem_ChkAlloc\r\n" ) ;
      return ( VDD_VMMem_ReAlloc( hVM, pVMCB, &pVMCB -> VDD_MStt ) ) ;
   }

} // end of VDD_VMMem_CA_CheckPlane0()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_CA_CheckPlanes
//
//  Description:
//     Checks the current size of the plane(s) and reallocates
//     if necessary.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     UINT uAllocNeeded
//        new size of plane(s)
//
//  Return Value:
//     TRUE if successful, FALSE otherwise
//
//  History:   Date       Author      Comment
//              4/29/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_CA_CheckPlanes
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uAllocNeeded
)
{
   int  i, nPlanesChgd = 0 ;

   for (i = 0; i < 4; i++)
   {
      if (pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_size <= uAllocNeeded)
      {
         pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_size = uAllocNeeded ;
         nPlanesChgd++ ;
      }
   }

   if (nPlanesChgd)
   {
      TRACEMSG( "Calling VDD_VMMem_ReAlloc from VDD_VMMem_ChkAlloc\r\n" ) ;
      return ( VDD_VMMem_ReAlloc( hVM, pVMCB, &pVMCB -> VDD_MStt ) ) ;
   }
   return ( TRUE ) ;

} // end of VDD_VMMem_CA_CheckPlanes()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_ChkAlloc
//
//  Description:
//     Check and adjust save memory allocation if necessary based
//     upon a faulting page id.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//     UINT uPageID
//        last page of plane
//
//  Return Value:
//     Returns FALSE if unable to allocate memory for this request.
//
//  History:   Date       Author      Comment
//              4/27/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_ChkAlloc
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uPageID
)
{
   UINT  uAllocNeeded = uPageID + 1 ;

   if (0 == pVMCB -> VDD_Flags & fVDD_NoSaveRes)
   {
      // Q: CGA compatibility enabled?

      if (pVMCB -> VDD_Stt.crtc.C_Mode & 1)
      {
         if ((fGrp6Char | fGrp6Chain) ==
               pVMCB -> VDD_Stt.G_Misc & (fGrp6Char | fGrp6Chain))
            return ( VDD_VMMem_CA_CheckPlanes( hVM, pVMCB, uAllocNeeded ) ) ;

         // Check for odd/even mode...

         if (0 == pVMCB -> VDD_Stt.S_MMode & fSeq4SqAd)
            return ( VDD_VMMem_CA_CheckPlane0( hVM, pVMCB, uAllocNeeded ) ) ;

         // Chain4 mode?

         if (0 == pVMCB -> VDD_Stt.S_MMode & fSeq4Chain4)
            return ( VDD_VMMem_CA_CheckPlanes( hVM, pVMCB, uAllocNeeded ) ) ;
      }
      return ( VDD_VMMem_CA_CheckPlane0( hVM, pVMCB, uAllocNeeded ) ) ;
   }

   return ( TRUE ) ;
   
} // end of VDD_VMMem_ChkAlloc()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_MemC_AboutToChange
//
//  Description:
//     Call VDD_VM_Mem_Disable for previous MemC owner VM, if
//     it was owned by a VM other than the Pseudo VM.  This
//     routine is called when changing focus and when entering
//     message mode.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//  History:   Date       Author      Comment
//              4/27/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_MemC_AboutToChange
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if ((hVM == hVidMemCVM) && (0 == pVMCB -> VDD_Flags & fVDD_MemCmods))
      return ;

   if (!hVidMemCVM)
      return ;

   if (hVidMemCVM == hVidMsgPseudoVM)
      return ;

   pVMCB = (PVDDCB) (hVidMemCVM + dwVidCBOff) ;

   // Disable any mapped memory and copy dirty bits

   VDD_VMMem_Disable( hVidMemCVM, pVMCB ) ;

} // end of VDD_VMMem_MemC_AboutToChange()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_Disable
//
//  Description:
//     Called by VDDState when memory controller state is changing.
//
//  Parameters:
//     DWORD hVM
//        previous owner VM handle, or NULL
//
//     PVDDCB pVMCB
//        previous owner VM's control block, or NULL
//
//  History:   Date       Author      Comment
//              4/24/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_Disable
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   int  i ;

   TRACEMSG( "Disable mem called...\r\n" ) ;

   // Perhaps this would be more efficient if it
   // scheduled an event to do the unmapping.

   if ((dwVidFlags & fVid_MsgInit) && hVM)
   {
      if (pVMCB -> VDD_TFlags & fVT_SysVMnot1stBank)
      {
         if (vmmTestSysVMHandle( hVM ))
         {
            vmmModifyPageBits( hVM, 0xA0, 16, ~P_AVAIL, 0, PG_HOOKED, 0 ) ;
            return ;
         }
      }

      VDD_PHMem_ClearMemCMods( hVM, pVMCB ) ;

      for (i = 0; i < pVMCB -> VDD_VideoPages; i++)
      {
         if (pVMCB -> VDD_PageFlags[ i ] & fPageSaveMem)
            VDD_VMMem_UnmapSaveMem( hVM, pVMCB, i ) ;
         else if (pVMCB -> VDD_PageMap[ i ])
            VDD_VMMem_UnmapPage( hVM, pVMCB, pVMCB -> VDD_PageMap[ i ] ) ;
      }
      VDD_VMMem_DisableNull( hVM, pVMCB ) ;
   }

} // end of VDD_VMMem_Disable()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_VMMem_VMCanUseSaveMem
//  
//  Description:
//     Returns TRUE if VM is in a text mode, a CGA graphics mode
//     or a packed picked (256 color) mode.
//  
//  Parameters:
//      PVDDCB pVMCB
//        VM's control block
//  
//  Return (BOOL):
//      TRUE if VM can use saveme
//  
//  History:   Date       Author      Comment
//             10/20/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_VMMem_VMCanUseSaveMem
(
    PVDDCB         pVMCB
)
{
   if ((pVMCB -> VDD_Stt.S_MMode & fSeq4SqAd) &&
       (pVMCB -> VDD_Stt.crtc.C_Mode & 1) &&
       (pVMCB -> VDD_Stt.S_MMode & fSeq4Chain4))
      return FALSE ;

   return (0 == (pVMCB -> VDD_Stt.G_Mode & 0x03)) ;

} // VDD_VMMem_VMCanUseSaveMem()

//------------------------------------------------------------------------
//  VOID VDD_PFault
//
//  Description:
//     VIDEO page faults occur in a number of cases.  For full screen VMs
//     they can occur when there is a mode change. After doing a few
//     checks to make sure that the mode change does not adversely
//     affect other VMs, the correct physical memory is remapped and
//     execution continues.  Full screen VMs can also page fault when
//     the application accesses more memory than has been set aside for
//     the VM, forcing the VM to run full screen, not background. For
//     windowed VMs, the same cases arise, but the action is a little
//     different.
//
//  Parameters:
//     DWORD hVM
//        handle of current VM
//
//     UINT uPageNum
//        page number of faulting page
//
//  Return Value:
//     PTE fixed or VM state modified so it won't run
//
//  History:   Date       Author      Comment
//              4/12/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID CDECL VDD_PFault
(
    DWORD           hVM,
    UINT            uPageNum
)
{
   BOOL    fAllocPage = TRUE ;
   DWORD   hPage ;
   PVDDCB  pVMCB ;
   UINT    uPageID, uPhysPage ;

   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;

   VDD_State_ScheduleDebugEvent( NULL ) ;

#ifdef DEBUG_verbose
   if (vmmTestSysVMHandle( hVM ))
      TRACEMSG( "p" ) ;
   else
      TRACEMSG( "P" ) ;
#endif
   vmmBeginCriticalSection( hVM,
                            Block_Svc_If_Ints_Locked | Block_Enable_Ints ) ;

   QUEUEMSG( "VDD pfault" ) ;

#ifdef SysVMin2ndBank
   if ((0 == (pVMCB -> VDD_TFlags & fVT_SysVMnot1stBank)) ||
       !vmmTestSysVMHandle( hVM ))
#endif
   {
      pVMCB -> VDD_Flags |= fVDD_PageFault | fVDD_InPFault ;
      if ((hVM == hVidCRTCVM) && (hVM == hVidMemCVM))
         VDD_State_SaveCRTCOwner( hVM, pVMCB ) ;
      if (!VDD_PHMem_CalcPageId( hVM, pVMCB, uPageNum, &uPageID ) ||
          ((uPageID >= pVMCB -> VDD_VideoPages) &&
           !VDD_VMMem_GrowPageMaps( pVMCB, uPageID )))
      {
         TRACEMSG( "VDD_PHMem_CalcPageId failed in PFault.\r\n" ) ;
         VDD_State_EndSRSection() ;
         VDD_VMMem_MapNull( hVM, pVMCB, uPageNum ) ;
         goto PF_Exit ;
      }

      if (0 == (pVMCB -> VDD_Flags & fVDD_Font))
      {
         // Check for font loading.  When loading a font without
         // using the INT 10h function, the controller state has to
         // be modified so that the 3rd memory plane can be accessed,
         // so the MemC state is modified for a pseudo planar state,
         // but the CRTC state is left unmodified.  So we check first
         // for a character cell height greater than 2, if it is 2 or
         // less, then we assume that the VM is in a normal graphics
         // mode and do no further checks.	If char cell hgt is
         // greater than 2, then we assume the VM is in a text mode
         // and check to see if the MemC looks right.  Some chipsets
         // (specifically the Tseng Labs ET3000) require the graphics
         // mode bit (0) in graphics controller miscellaneous register
         // (index 6) to be set, indicating graphics mode to load the
         // font, so we check this bit first; and if it is set, then we
         // assume that a font is being loaded.  Otherwise we just
         // check to see if the graphics controller is still in chained
         // odd to even mode by checking bit 1 in the graphics controller
         // miscellaneous register; if not, then we assume that a font
         // is being loaded, because fonts are stored in consecutive
         // bytes in plane 3.

         if (((pVMCB -> VDD_Stt.crtc.C_CharHgt & 0x1F) > 2) ||
             (pVMCB -> VDD_Stt.G_Misc & fGrp6Char) ||
             (0 == pVMCB -> VDD_Stt.G_Misc & fGrp6Chain))
         {
            TRACEMSG( "font loading detected in VDD_PFault\r\n" ) ;
            pVMCB -> VDD_Flags |= fVDD_DirFont ;

            pVMCB -> VDD_fontpages |= (1 << (uPageID >> 1)) ;
         }
      }
      else
         pVMCB -> VDD_fontpages |= (1 << (uPageID >> 1)) ;

      // Check to see if we can map regular save memory into the VM.
      // The VM must be in a text, CGA graphics or 256 color graphics
      // mode (none of which are planar modes.)
      // 
      // For efficiency of CRTC VM, we give the VM physical memory,
      // even though we could treat this as virtual video page
      
      if (hVM != hVidCRTCVM)
      {
         // waiting to be attached? if so, just map null mem

         if (pVMCB -> VDD_Flags & fVDD_WaitAtt)
         {
            VDD_State_EndSRSection() ;
            VDD_VMMem_MapNull( hVM, pVMCB, uPageNum ) ;
            goto PF_Exit ;
         }

         if ((pVMCB -> VDD_Flags & (fVDD_Font | fVDD_DirFont)) ||
             VDD_VMMem_VMCanUseSaveMem( pVMCB ))
         {
            if (VDD_VMMem_MapSaveMem( hVM, pVMCB, uPageID, uPageNum ))
               goto PF_Exit ;

            // otherwise fall through...
         }

      }

      // Must use the video memory...

      if (!(hPage = pVMCB -> VDD_PageMap[ uPageID ]))
      {
         if (0 == (pVMCB -> VDD_Flags & fVDD_ModeSet))
            VDD_VMMem_ChkAlloc( hVM, pVMCB, uPageID ) ;

         // falls through to allocate page...
      }
      else
      {
         // Have page...

         if (hVM == hVidCRTCVM)
         {
            if (!VDD_PHMem_MemMappedPhys( hVM, pVMCB, hPage, uPageID ))
            {
               // Free current page and allocate correct page...

               VDD_State_BeginSRSection() ;
               VDD_PHMem_FreePage( hPage ) ;
            }
            else
               fAllocPage = FALSE ;
         }
      }

      // Allocate page...

      if (fAllocPage)
      {
         VDD_State_BeginSRSection() ;
         dwVidUnaccessedPage = 0 ;
         if (hPage =
                VDD_PHMem_AllocVideoPage( hVM, pVMCB, uPageID,
                                          VDD_VMMem_PageHandler ))
         {
            if (dwVidUnaccessedPage)
               vmmScheduleEvent( pVDD_Error_CvtFullScrn ) ;
         }

         // Did we fail to allocate?

         if (!hPage)
         {
            VDD_State_EndSRSection() ;
            TRACEMSG( "VDD_PHMem_AllocVideoPage failed.\r\n" ) ;
            if (pVMCB -> VDD_Flags & fVDD_CanRestore)
            {
               // Supposed to notify?

               if (pVMCB -> VDD_TFlags & fVT_Notify_at_PF)
                  VDD_Proc_BGrndNotify( hVM, pVMCB ) ;
            }
            else
            {
               // We're here because the VM can't allocate a page
               // of video memory.  We're going to map a NUL page
               // so it is possible that the VM's display will be
               // corrupt.  If the VM is doing a mode change, then
               // we will assume that this is just the process of
               // initializing memory, so we won't give an error
               // message, otherwise...

               if (0 == (pVMCB -> VDD_Flags & fVDD_ModeSet) &&
                   0 == (pVMCB -> VDD_Flags & fVT_SuprsMsg))
                     VDD_Error_NoPagesAvail() ;
            }

            VDD_VMMem_MapNull( hVM, pVMCB, uPageNum ) ;
            goto PF_Exit ;
         }
         else
         {
            pVMCB -> VDD_PageMap[ uPageID ] = hPage ;
            pVMCB -> VDD_PageFlags[ uPageID ] &= fPageGblFlags ;
            VDD_VMMem_RestorePage( hVM, pVMCB, hPage,
                                   &pVMCB -> VDD_PageFlags[ uPageID ] ) ;
         }
      }

      // Map the page...

      if ((hVM == hVidCRTCVM) && 0 == (pVMCB -> VDD_Flags & fVDD_ModeSet))
      {
         UINT  uVisFirst, uVisPages ;

         VDD_PHMem_GetVisiblePages( hVM, pVMCB, &uVisFirst, &uVisPages ) ;
         if ((uVisPages != pVMCB -> VDD_visible_pages) ||
             (uVisFirst != pVMCB -> VDD_first_page))
         {
            TRACEMSG( "Adjusting visible pages...\r\n" ) ;

            VDD_VMMem_AdjustVisiblePages( hVM, pVMCB ) ;
         }
      }
      VDD_State_EndSRSection() ;
      uPhysPage = VDD_PHMem_AccessPage( hVM, pVMCB, hPage, 1 ) ;

      if (0 != ((PPAGEDATA) hPage)  -> PDS_linear_page)
         VDD_VMMem_UnmapPage( hVM, pVMCB, hPage ) ;

      ((PPAGEDATA) hPage) -> PDS_linear_page = uPageNum ;
      vmmPhysIntoV86( uPhysPage, hVM, uPageNum, 1, 0 ) ;

PF_Exit:
      pVMCB -> VDD_Flags &= ~(fVDD_DirFont | fVDD_InPFault) ;
      vmmEndCriticalSection( hVM ) ;

   }
#ifdef SysVMin2ndBank
   else
   {
      QUEUEMSG( "VDD Sys VM pfault\r\n" ) ;
      if (hVM != hVidMemCVM)
         VDD_State_SetMemCOwner( hVM, pVMCB ) ;

      if (uPageNum < 0xB0)
      {
         // Map all 16 pages...

         vmmPhysIntoV86( 0xA0, hVM, 0xA0, 16, 0 ) ;
      }
      else
      {
         // round down to B0 or B8 and map in 8
         uPageNum &= ~0x07 ;
         vmmPhysIntoV86(  uPageNum, hVM, uPageNum, 8, 0 ) ;
      }
      vmmEndCriticalSection( hVM ) ;
   }
#endif

} // end of VDD_PFault()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_VMMem_MapSaveMem
//  
//  Description:
//     Directly map a page of the VM's video save memory into
//     the VM.  This routine is called by VDD_PFault for VM's
//     which can run in the background without using demand
//     paged video memory.  Text, CGA graphics and 256 color
//     graphics modes all operate with a single logical plane
//     of memory, so we can just map normal memory into the
//     required video space and the VM can modify it directly.
//  
//  Parameters:
//      DWORD hVM
//  
//      PVDDCB pVMCB
//  
//      UINT uPageID
//  
//      UINT uPageNum
//  
//  Return (BOOL):
//      FALSE if page not available in save memory
//  
//  History:   Date       Author      Comment
//             10/21/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_VMMem_MapSaveMem
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uPageID,
    UINT            uPageNum
)
{

   DWORD          hPage ;
   PMSSPLANEINFO  pMPI ;

   pMPI = &pVMCB -> VDD_MStt.MSS_planes[ 0 ] ;

   // Mapping save memory for font load???

   if (pVMCB -> VDD_Flags & (fVDD_Font | fVDD_DirFont))
   {
      // Map plane 2

      pMPI += 2 ;

      if (uPageID >= pMPI -> MSS_plane_size)
      {
         // two pages for font, so force to send
         // and convert page id to # of pages

         pMPI -> MSS_plane_size = (uPageID | 0x01) + 1 ;
         goto MSM_AttemptAlloc ;
      }
   }
   else
   {
      if (uPageID >= pMPI -> MSS_plane_size)
      {
#ifdef DEBUG_verbose
         TRACEMSG( "not enough save memory to map for page\r\n" ) ;
         return FALSE ;
#endif
      }
   }

   if (uPageID >= pMPI -> MSS_plane_alloced)
   {
MSM_AttemptAlloc:
      if (!VDD_VMMem_ReAllocPlane( hVM, pMPI ))
      {
         // Not critical if not loading font

         if (pVMCB -> VDD_Flags & (fVDD_Font | fVDD_DirFont))
            return TRUE ;

         // We couldn't allocate save memory at this time, and since this
         // is for loading a font, we don't want to proceed with null
         // memory, because the font will be lost.  So, we do a critical
         // set focus to convert the VM to full screen.  Now the VM will
         // own physical video memory, and we can just return a failure
         // from this routine.  VDD_Pfault will see the failure, but will
         // also see that the VM owns a physical page so it will
         // just use it.

         // VM is converted to full screen as a desired side effect of
         // SHELL_SYSMODAL_Message.

         VDD_Error_CvtFullScrn( hVM, NULL, NULL ) ;
         return FALSE ;
      }
   }

   // Force a page lock in case we have an intelligent pageswap
   // device, in which case the mmgr won't do an implied lock.

   if (!vmmPageLock( pMPI -> MSS_plane_handle, 1, uPageID, 0 ))
   {
      DEBUGMSG( "page lock on save memory failed...\r\n" ) ;
      return FALSE ;
   }

   if (!vmmMapIntoV86( pMPI -> MSS_plane_handle, hVM, uPageID,
                       1, uPageNum, 0 ))
   {
#ifdef DEBUG
      DEBUGMSG( "MapIntV86 failed in VDD_VMMem_MapSaveMem\r\n" ) ;
      return FALSE ;
#endif
   }

   pVMCB -> VDD_savemem_maps += 1 ;

   if (hPage = pVMCB -> VDD_PageMap[ uPageID ])
      VDD_PHMem_FreePage( hPage ) ;

   pVMCB -> VDD_PageFlags[ uPageID ] |= fPageSaveMem ;
   if (pVMCB -> VDD_Flags & (fVDD_Font | fVDD_DirFont))
      pVMCB -> VDD_PageFlags[ uPageID ] |= fPageFontMem ;

   return TRUE ;

} // VDD_VMMem_MapSaveMem()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_UnmapSaveMem
//
//  Description:
//     Copy dirty bit status from the VM's page table into the
//     VM's VDD_PageFlags table.  Then, unmaps the page.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//     UINT uPageID
//        page ID
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_UnmapSaveMem
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uPageID
)
{
   DWORD          dwPTE ;
   UINT           uPageNum, uBiasPageNum ;
   PMSSPLANEINFO  pMPI ;

   uPageNum = VDD_PHMem_CalcPageAddr( hVM, pVMCB, uPageID ) ;
   uBiasPageNum =
      ((uPageNum << 12) + ((PVMMCB) hVM) -> CB_High_Linear) >> 12 ;
   vmmCopyPageTable( uBiasPageNum, 1, &dwPTE, 0 ) ;
   if (dwPTE & P_DIRTY)
   {
      pVMCB -> VDD_PageFlags[ uPageID ] |= fPageSaved ;
      VDD_VMMem_RecordVRAMMod( pVMCB, uPageID ) ;
   }

   pVMCB -> VDD_PageFlags[ uPageID ] &= ~fPageSaveMem ;
   pVMCB -> VDD_savemem_maps-- ;

   vmmModifyPageBits( hVM, uPageNum, 1, ~P_AVAIL, 0, PG_HOOKED, 0 ) ;

   // Default to plane 0

   pMPI = &pVMCB -> VDD_MStt.MSS_planes[ 0 ] ;

   if (pVMCB -> VDD_PageFlags[ uPageID ] & fPageFontMem)
   {
      // If font page, index to font page

      pVMCB -> VDD_PageFlags[ uPageID ] &= ~fPageFontMem ;
      pMPI = &pVMCB -> VDD_MStt.MSS_planes[ 2 ] ;
   }
   if (uPageID < pMPI -> MSS_plane_alloced)
      vmmPageUnLock( pMPI -> MSS_plane_handle, 1, uPageID, 0 ) ;

} // end of VDD_VMMem_UnmapSaveMem()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_MapNull
//
//  Description:
//     Maps null page at uPageNum.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//     UINT uPageNum
//        page number
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_MapNull
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uPageNum
)
{
   pVMCB -> VDD_nullpages |= 1 << (uPageNum - 0xA0) ;
   vmmMapIntoV86( vmmGetNulPageHandle(), hVM, uPageNum, 1, 0, 0 ) ;

} // end of VDD_VMMem_MapNull()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_DisableNull
//
//  Description:
//     Unmap all pages mapped into a VM.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_DisableNull
(
    DWORD               hVM,
    PVDDCB              pVMCB
)
{
   UINT  uPage, uNullPages ;

   TRACEMSG( "disable null mem\r\n" ) ;

   if (pVMCB -> VDD_nullpages)
   {
      uPage = 0xA0 ;
      uNullPages = pVMCB -> VDD_nullpages ;
      for (uPage = 0xA0; uNullPages != 0; uPage++, uNullPages >>= 1)
         if (uNullPages & 1)
            vmmModifyPageBits( hVM, uPage, 1, ~P_AVAIL, 0, PG_HOOKED, 0 ) ;
   }
   pVMCB -> VDD_nullpages = 0 ;

} // end of VDD_VMMem_DisableNull()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_PageHandler
//
//  Description:
//     Handler for all demand paged video pages owned by VMs.
//     It handles tracking a page's dirty bit, saving the memory
//     contents to the VM's save memory and unmapping the page
//     as required.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     DWORD dwNotify
//        notification id
//           0 - page is about to be freed
//           1 - page is no longer accessible with the
//               current memory access state
//           2 - physical page assignment is changing
//
//     DWORD hPage
//        page handle
//
//     DWORD hNewPage
//        new page handle if dwNotify == 2
//
//  Return Value:
//     TRUE if dwNotify != 2, otherwise FALSE if ECX == 2 to indicate
//        that page assignment changing is not supported
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_PageHandler
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    DWORD           dwNotify,
    DWORD           hPage,
    DWORD           hNewPage
)
{
   UNUSED_PARAM( hNewPage ) ;

   switch (dwNotify)
   {
      case 0:
      {
         BYTE   fSaved ;
         UINT   uPageID ;

         fSaved = 0 ;
         if (PGS_DIRTY == VDD_VMMem_IsPageDirty( hVM, hPage ))
         {
            VDD_VMMem_SavePage( hVM, pVMCB, hPage ) ;
            fSaved = fPageSaved ;

            uPageID = ((PPAGEDATA) hPage) -> PDS_virtual_id ;
            VDD_VMMem_RecordVRAMMod( pVMCB, uPageID ) ;
         }
         uPageID = ((PPAGEDATA) hPage) -> PDS_virtual_id ;

#ifdef DEBUG
         if (uPageID >= pVMCB -> VDD_VideoPages)
            DEBUGMSG( "VDD: bad virtual id\r\n" ) ;
         if (pVMCB -> VDD_PageMap[ uPageID ] != hPage)
            DEBUGMSG( "VDD: clearing invalid page handle entry\r\n" ) ;
#endif
         // Clear page handle and owned bit

         pVMCB -> VDD_PageMap[ uPageID ] = 0 ;
         pVMCB -> VDD_PageFlags[ uPageID ] &= fPageGblFlags ;
         pVMCB -> VDD_PageFlags[ uPageID ] |= fSaved ;

         // Set SHADOW_TRASHED bit in display driver if necessary

         if (pVMCB -> VDD_Flags & fVDD_DspDrvrAware)
         {
            PBYTE  pData ;

            if (HIWORD( dwVidShadowMemStatusPtr ))
            {
               pData =
                  (PBYTE)
                     vmmSelectorMapFlat( hVM,
                                         HIWORD( dwVidShadowMemStatusPtr ),
                                         0 ) ;
               if (-1 != pData)
               {
                  pData += LOWORD( dwVidShadowMemStatusPtr ) ;
                  if (*pData & SHADOW_IN_USE)
                     *pData |= SHADOW_TRASHED ;
               }
            }
         }
      }

      // ... fall-through ...

      case 1:
         VDD_VMMem_UnmapPage( hVM, pVMCB, hPage ) ;
         break ;

      case 2:
         return ( FALSE ) ;

      default:
         DEBUGMSG( "VDD_VMMem_PageHandler unknown page notification\r\n" ) ;
   }
   return ( TRUE ) ;

} // end of VDD_VMMem_PageHandler()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_VMMem_Msg_PageHandler
//  
//  Description:
//      Returns TRUE.
//  
//  Parameters:
//      DWORD hVM
//  
//      PVDDCB pVMCB
//  
//      DWORD dwNotify
//  
//      DWORD hPage
//  
//      DWORD hNewPage
//  
//  Return (BOOL):
//      TRUE
//  
//  History:   Date       Author      Comment
//             10/21/93   BryanW      
//  
//--------------------------------------------------------------------------

BOOL VDD_VMMem_Msg_PageHandler
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    DWORD           dwNotify,
    DWORD           hPage,
    DWORD           hNewPage
)
{
   UNUSED_PARAM( hVM ) ;
   UNUSED_PARAM( pVMCB ) ;
   UNUSED_PARAM( dwNotify ) ;
   UNUSED_PARAM( hPage ) ;
   UNUSED_PARAM( hNewPage ) ;

   return TRUE ;

} // VDD_VMMem_Msg_PageHandler()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VMMem_UnmapPage
//  
//  Description:
//      Unmap a physical video page from a VM.  ORing the current
//      state of physical dirty bit into the pageflags entry for
//      the virtual page.
//  
//  Parameters:
//      DWORD hVM
//  
//      PVDDCB pVMCB
//  
//      DWORD hPage
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/23/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VMMem_UnmapPage
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    DWORD           hPage
)
{
   DWORD  dwLinPage ;
   DWORD  dwPgStatus ;

#ifdef DEBUG_verbose
   QUEUEMSG( "Unmap page called.\r\n" ) ;
#endif
   if (dwLinPage = ((PPAGEDATA) hPage) -> PDS_linear_page)
   {
      // Don't check accessed if not pfault

      if (pVMCB -> VDD_Flags & fVDD_InPFault)
      {
         // Does page belong to current VM and
         // was it accessed?

         if (vmmTestCurVMHandle( hVM ))
         {
            DWORD dwPTE ;

            dwPTE = VDD_VMMem_GetPTE( hVM, dwLinPage ) ;
            if (0 == (dwPTE & P_ACC))
               dwVidUnaccessedPage = hPage ;
         }
      }

      if (PGS_NOTMAPPED !=
            (dwPgStatus = VDD_VMMem_IsPageDirty( hVM, hPage )))
      {
         if (PGS_DIRTY == dwPgStatus)
            VDD_VMMem_RecordVRAMMod( pVMCB,
                                     ((PPAGEDATA) hPage) -> PDS_virtual_id ) ;
         ((PPAGEDATA) hPage) -> PDS_linear_page = 0 ;

         VDD_PHMem_ReleasePage( hPage ) ;
         vmmModifyPageBits( hVM, dwLinPage, 1, ~P_AVAIL, NULL, PG_HOOKED,
                            NULL ) ;
      }
   }

} // VDD_VMMem_UnmapPage()

//------------------------------------------------------------------------
//  DWORD VDD_VMMem_GetPTE
//
//  Description:
//     Get the PTE for the given linear page.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     UINT uLinPageNum
//        linear page number
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

DWORD VDD_VMMem_GetPTE
(
    DWORD           hVM,
    UINT            uLinPageNum
)
{
   DWORD  dwPTE ;
   UINT   uBiasPageNum ;

   uBiasPageNum =
      ((uLinPageNum << 12) + ((PVMMCB) hVM) -> CB_High_Linear) >> 12 ;
   vmmCopyPageTable( uBiasPageNum, 1, &dwPTE, 0 ) ;
   return ( dwPTE ) ;

} // end of VDD_VMMem_GetPTE()

//--------------------------------------------------------------------------
//  
//  DWORD VDD_VMMem_IsPageDirty
//  
//  Description:
//      Or page table dirty bit into page handle's flags, and
//      check to see if dirty bit is set in page handle's flags.
//  
//  Parameters:
//      DWORD hVM
//  
//      DWORD hPage
//  
//  Return (DWORD):
//     PGS_NOTMAPPED if page is not mapped
//     PGS_DIRTY if page is dirty
//        otherwise 0
//  
//  History:   Date       Author      Comment
//             10/23/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

DWORD VDD_VMMem_IsPageDirty
(
    DWORD           hVM,
    DWORD           hPage
)
{
   DWORD dwLinPage, dwPTE ;

   if (0 == (dwLinPage = ((PPAGEDATA) hPage) -> PDS_linear_page))
      return PGS_NOTMAPPED ;

   if (0 == (((PPAGEDATA) hPage) -> PDS_flags & fPDS_Dirty))
      return 0 ;

   dwPTE = VDD_VMMem_GetPTE( hVM, dwLinPage ) ;
   if (dwPTE & P_DIRTY)
      ((PPAGEDATA) hPage) -> PDS_flags |= P_DIRTY ;

   if (0 == (((PPAGEDATA) hPage) -> PDS_flags & fPDS_Dirty))
      return 0 ;

   return PGS_DIRTY ;

} // VDD_VMMem_IsPageDirty()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_VMMem_ChkSaveMem
//  
//  Description:
//      Try to alloc required save mem for VM, if previously
//      unsuccessful.
//
//  Parameters:
//      DWORD hVM
//  
//      PVDDCB pVMCB
//  
//  Return (BOOL):
//      FALSE if no memory available.
//
//  History:   Date       Author      Comment
//             10/23/93   BryanW      
//  
//--------------------------------------------------------------------------


BOOL VDD_VMMem_ChkSaveMem
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   UINT       i ;  
   PMEMSTATE  pMS ;

   if (pVMCB -> VDD_EFlags & fVDE_NoMain)
   {
      pMS = &pVMCB -> VDD_MStt ;
      for (i = 0; i < 4; i++)
      {
         if (pMS -> MSS_planes[ i ].MSS_plane_size >
                pMS -> MSS_planes[ i ].MSS_plane_alloced)
            if (!VDD_VMMem_ReAllocPlane( hVM, &pMS -> MSS_planes[ i ] ))
               return FALSE ;
      }
      pVMCB -> VDD_EFlags &= ~fVDE_NoMain ;
   }
   return TRUE ;

} // VDD_VMMem_ChkSaveMem()

//--------------------------------------------------------------------------
//  
//  VOID MemC_UpdateReg
//  
//  Description:
//      Support routine for VDD_VMMem_ChkMemCState().
//  
//  Parameters:
//      WORD wPort
//  
//      DWORD dwValue
//  
//  Return (VOID):
//  
//  History:   Date       Author      Comment
//             10/23/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID MemC_UpdateReg
(
    WORD            wPort,
    DWORD           dwValue
)
{
   BYTE  bValue ;

   // dwValue = or mask | and mask | index

   outp( wPort, (BYTE) dwValue ) ;
   bValue = inp( wPort + 1 ) ;
   bValue &= (BYTE) (dwValue >> 8) ;
   bValue |= (BYTE) (dwValue >> 16) ;
   outp( wPort + 1, bValue ) ;

   // flag that MemC is owned by VDD

   VDD_State_ClearMemCOwner() ;

} // MemC_UpdateReg()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VMMem_ChkMemCState
//  
//  Description:
//     Update the MemC state if necessary to allow reading &
//     writing video memory for save/restore routines.
//
//  Parameters:
//      PVDDCB pVMCB
//         pointer to VM's control block
//  
//      BYTE bMxPlane
//         write plane mask for sequencer
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/23/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_VMMem_ChkMemCState
(
    PVDDCB          pVMCB,
    BYTE            bMxPlane
)
{
   WORD  wPort ;

   // Graphics controller

   wPort = 0x3CE ;
   if (pVMCB -> VDD_Stt.G_SREna)
      MemC_UpdateReg( wPort, 0x00000001 ) ;

   if (pVMCB -> VDD_Stt.G_Func)
      MemC_UpdateReg( wPort, 0x00000003 ) ;

   if (pVMCB -> VDD_Stt.G_Mode & 0x0B)
      MemC_UpdateReg( wPort, 0x0000F405 ) ;

   if (-1 == pVMCB -> VDD_Stt.G_Mask)
      MemC_UpdateReg( wPort, 0x00FF0008 ) ;

   // Sequencer

   wPort = 0x3C4 ;

   if (bMxPlane != pVMCB -> VDD_Stt.S_Mask)
      MemC_UpdateReg( wPort, (((DWORD) bMxPlane) << 16) | 0x00000002 ) ;

} // VDD_VMMem_ChkMemCState()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_VMMem_SavePage
//  
//  Description:
//      Copy a video page's data into the VM's copy memory, based
//      on the current mode of memory access for the VM.
//
//  Parameters:
//      DWORD hVM
//  
//      PVDDCB pVMCB
//  
//      DWORD hPage
//  
//  Return (BOOL):
//  
//  History:   Date       Author      Comment
//             10/23/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_VMMem_SavePage
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    DWORD           hPage
)
{
   UINT  i, uPhysPage ;

   vmmBeginCriticalSection( hVM,
                            Block_Svc_If_Ints_Locked | Block_Enable_Ints ) ;

   QUEUEMSG( "VDD: save page...\r\n" ) ;

   ((PPAGEDATA) hPage) -> PDS_flags &= ~fPDS_Dirty ;

   // Try to fill out save memory, if not already allocated.

   VDD_VMMem_ChkSaveMem( hVM, pVMCB ) ;

   // Tell VDDSTATE that the latches are about to be modified

   if (hVM == hVidMemCVM)
      VDD_State_SetLatchesOwner( NULL, NULL, NULL ) ;

   VDD_State_SetMemCOwner( hVM, pVMCB ) ;

   if (0 == (pVMCB -> VDD_Stt.crtc.C_Mode & 1))
   {
SP_OddEven:

      // Copy planes 0 & 1 as 1 plane and then copy
      // the font in plane 2.

      uPhysPage = VDD_PHMem_AccessPage( hVM, pVMCB, hPage, 0 ) ;

      VDD_VMMem_ChkMemCState( pVMCB, 3 ) ;
      VDD_VMMem_CopyFromPlane( hVM, &pVMCB -> VDD_MStt.MSS_planes[ 0 ],
                               hPage, uPhysPage ) ;

      // Copy font if not in graphics mode.

      if (0 == (pVMCB -> VDD_Stt.G_Misc & fGrp6Char))
      {
         VDD_State_SetMemCPlanar() ;

         // flag pages as planar mode

         uPhysPage = VDD_PHMem_AccessPage( 0, pVMCB, hPage, 0 ) ;

         VDD_VMMem_SelectPlane( 2, 4 ) ;

         VDD_Font_SaveFont( hVM, pVMCB,  &pVMCB -> VDD_MStt.MSS_planes[ 2 ],
                            hPage, uPhysPage ) ;

         VDD_State_SetMemCOwner( hVM, pVMCB ) ;
      }
      QUEUEMSG( "VDD: end save page...\r\n" ) ;
      vmmEndCriticalSection( hVM ) ;
      return TRUE ;
   }

   if ((fGrp6Char | fGrp6Chain) ==
       (pVMCB -> VDD_Stt.G_Misc & (fGrp6Char | fGrp6Chain)))
   {
      // chained graphics mode

      uPhysPage = VDD_PHMem_AccessPage( hVM, pVMCB, hPage, 0 ) ;

      // modify state to ensure that we're reading planes 0 & 1

      VDD_State_ClearMemCOwner() ;
      VDD_VMMem_ChkMemCState( pVMCB, 3 ) ;

      // reads from planes 0 & 1

      VDD_VMMem_SelectPlane( 0, 0 ) ;

      VDD_VMMem_CopyFromPlane( hVM, &pVMCB -> VDD_MStt.MSS_planes[ 0 ],
                               hPage, uPhysPage ) ;

      VDD_VMMem_SelectPlane( 2, 0 ) ;

      VDD_VMMem_CopyFromPlane( hVM, &pVMCB -> VDD_MStt.MSS_planes[ 2 ],
                               hPage, uPhysPage ) ;

      QUEUEMSG( "VDD: end save page...\r\n" ) ;
      vmmEndCriticalSection( hVM ) ;
      return TRUE ;
   }

   if (pVMCB -> VDD_Stt.S_MMode & fSeq4Chain4)
   {
      // chain4 mode - copy one plane

      uPhysPage = VDD_PHMem_AccessPage( hVM, pVMCB, hPage, 0 ) ;

      VDD_VMMem_ChkMemCState( pVMCB, 0x0F ) ;
      VDD_VMMem_CopyFromPlane( hVM, &pVMCB -> VDD_MStt.MSS_planes[ 0 ],
                               hPage, uPhysPage ) ;

      QUEUEMSG( "VDD: end save page...\r\n" ) ;

      vmmEndCriticalSection( hVM ) ;
      return TRUE ;
   }
   else if (pVMCB -> VDD_Stt.S_MMode & fSeq4SqAd)
   {
      PMEMSTATE  pMS ;

      // not chain4 - copy all four planes

      VDD_State_SetMemCPlanar() ;

      // flag pages as planar mode

      uPhysPage = VDD_PHMem_AccessPage( 0, pVMCB, hPage, 0 ) ;

      pMS = &pVMCB -> VDD_MStt ;
      for (i = 0; i < 4; i++)
      {
         VDD_VMMem_SelectPlane( i, 0 ) ;

         if ((pMS -> MSS_planes[ i ].MSS_plane_size != 0) || (i == 1))
            VDD_VMMem_CopyFromPlane( hVM, &pMS -> MSS_planes[ i ],
                                     hPage, uPhysPage ) ;
         else
         {
            // we are probably saving a text page while the VM is
            // temporarily in a planar mode while it is loading a
            // soft font, so we need to save the attributes into
            // plane 0 save mem.

            if (0 == (pVMCB -> VDD_Flags & (fVDD_Font | fVDD_DirFont)))
            {
               DEBUGMSG( "saving plane 1 with no save mem and not loading font!\r\n" ) ;
               VDD_VMMem_CopyFromPlane( hVM, &pMS -> MSS_planes[ i ],
                                        hPage, uPhysPage ) ;
            }
            else
               VDD_VMMem_SaveAttributePage( hVM, &pMS -> MSS_planes[ 0 ],
                                            hPage, uPhysPage ) ;
         }
      }
      VDD_State_SetMemCOwner( hVM, pVMCB ) ;

      QUEUEMSG( "VDD: end save page...\r\n" ) ;

      vmmEndCriticalSection( hVM ) ;
      return TRUE ;
   }
   else
      goto SP_OddEven ;

} // VDD_VMMem_SavePage()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_VMMem_RestorePage
//  
//  Description:
//      Restore a video page's data from the VM's copy memory,
//      based on the current mode of memory access for the VM.
//  
//  Parameters:
//      DWORD hVM
//  
//      PVDDCB pVMCB
//  
//      DWORD hPage
//  
//      PBYTE pPageFlags
//  
//  Return (BOOL):
//      TRUE if successful
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_VMMem_RestorePage
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    DWORD           hPage,
    PBYTE           pPageFlags
)
{
   UINT   i, uPhysPage ;
   DWORD  pVDD_RestoreGraphics,
          pVDD_RestoreText,
          pVDD_RestoreFont ;

   vmmBeginCriticalSection( hVM,
                            Block_Svc_If_Ints_Locked | Block_Enable_Ints ) ;

   QUEUEMSG( "VDD: restore page...\r\n" ) ;

   if (*pPageFlags & fPageSaved)
   {
      pVDD_RestoreGraphics = (DWORD) VDD_VMMem_CopyToPlane ;
      pVDD_RestoreText = (DWORD) VDD_VMMem_CopyToPlane ;
      pVDD_RestoreFont = (DWORD) VDD_Font_RestoreFont ;
   }
   else
   {
      pVDD_RestoreGraphics = (DWORD) VDD_VMMem_ZeroPage ;
      pVDD_RestoreText = (DWORD) VDD_VMMem_InitText ;
      pVDD_RestoreFont = (DWORD) VDD_VMMem_ZeroPage ;

   }

   VDD_State_SetMemCOwner( hVM, pVMCB ) ;
   if (0 == (pVMCB -> VDD_Stt.crtc.C_Mode & 1))
   {
RP_OddEven:
      // odd/even mode
      
      // copy planes 0 & 1 as 1 plane and then copy
      // the font in plane 2

      uPhysPage = VDD_PHMem_AccessPage( hVM, pVMCB, hPage, 0 ) ;

      // enable writes to planes 0 & 1

      VDD_VMMem_ChkMemCState( pVMCB, 3 ) ;

      if (0 == (pVMCB -> VDD_Stt.G_Misc & fGrp6Char))
      {
         if (pVDD_RestoreText == (DWORD) VDD_VMMem_InitText)
            VDD_VMMem_InitText( uPhysPage, 0x2007 ) ;
         else
            VDD_VMMem_CopyToPlane( hVM, &pVMCB -> VDD_MStt.MSS_planes[ 0 ],
                                   hPage, uPhysPage ) ;

         // restore font if necessary 

         if (pVMCB -> VDD_fontpages &
               (1 << (((PPAGEDATA) hPage) -> PDS_virtual_id >> 1)))
         {
            VDD_State_SetMemCPlanar() ;

            // flag as planar mode

            uPhysPage = VDD_PHMem_AccessPage( 0, pVMCB, hPage, 0 ) ;

            VDD_VMMem_SelectPlane( 2, 4 ) ;

            if (pVDD_RestoreFont == (DWORD) VDD_VMMem_ZeroPage)
               VDD_VMMem_ZeroPage( uPhysPage ) ;
            else
               VDD_Font_RestoreFont( hVM, pVMCB,
                                     &pVMCB -> VDD_MStt.MSS_planes[ 2 ],
                                     hPage, uPhysPage ) ;

         }
      }
      else
      {
         // restore CGA graphics page

         if (pVDD_RestoreGraphics == (DWORD) VDD_VMMem_ZeroPage)
            VDD_VMMem_ZeroPage( uPhysPage ) ;
         else
            VDD_VMMem_CopyToPlane( hVM, &pVMCB -> VDD_MStt.MSS_planes[ 0 ],
                                   hPage, uPhysPage ) ;
      }

      QUEUEMSG( "VDD: end restore page...\r\n" ) ;

      vmmEndCriticalSection( hVM ) ;
      return TRUE ;
   }

   if ((fGrp6Char | fGrp6Chain) ==
         (pVMCB -> VDD_Stt.G_Misc & (fGrp6Char | fGrp6Chain)))
   {
      // chained graphics mode

      uPhysPage = VDD_PHMem_AccessPage( hVM, pVMCB, hPage, 0 ) ;

      // modify state to ensure that we're reading planes 0 & 1

      VDD_State_ClearMemCOwner() ;
      VDD_VMMem_ChkMemCState( pVMCB, 3 ) ;

      // reads & writes for planes 0 & 1

      VDD_VMMem_SelectPlane( 0, 3 ) ;

      if (pVDD_RestoreGraphics == (DWORD) VDD_VMMem_ZeroPage)
         VDD_VMMem_ZeroPage( uPhysPage ) ;
      else
         VDD_VMMem_CopyToPlane( hVM, &pVMCB -> VDD_MStt.MSS_planes[ 0 ],
                                hPage, uPhysPage ) ;

      // reads & writes for planes 2 & 3

      VDD_VMMem_SelectPlane( 2, 0xC ) ;

      if (pVDD_RestoreGraphics == (DWORD) VDD_VMMem_ZeroPage)
         VDD_VMMem_ZeroPage( uPhysPage ) ;
      else
         VDD_VMMem_CopyToPlane( hVM, &pVMCB -> VDD_MStt.MSS_planes[ 2 ],
                                hPage, uPhysPage ) ;

      QUEUEMSG( "VDD: end restore page...\r\n" ) ;

      vmmEndCriticalSection( hVM ) ;
      return TRUE ;
   }

   if (pVMCB -> VDD_Stt.S_MMode & fSeq4Chain4)
   {
      // chain4 mode

      uPhysPage = VDD_PHMem_AccessPage( hVM, pVMCB, hPage, 0 ) ;

      // enable writes to all planes

      VDD_VMMem_ChkMemCState( pVMCB, 0x0F ) ;

      // copy 1 plane

      if (pVDD_RestoreGraphics == (DWORD) VDD_VMMem_ZeroPage)
         VDD_VMMem_ZeroPage( uPhysPage ) ;
      else
         VDD_VMMem_CopyToPlane( hVM, &pVMCB -> VDD_MStt.MSS_planes[ 0 ],
                                hPage, uPhysPage ) ;

      QUEUEMSG( "VDD: end restore page...\r\n" ) ;

      vmmEndCriticalSection( hVM ) ;
      return TRUE ;
   }
   else if (pVMCB -> VDD_Stt.S_MMode & fSeq4SqAd)
   {
      PMEMSTATE  pMS ;

      // not chain4 - copy all four planes

      VDD_State_SetMemCPlanar() ;

      // flag pages as planar mode

      uPhysPage = VDD_PHMem_AccessPage( 0, pVMCB, hPage, 0 ) ;

      pMS = &pVMCB -> VDD_MStt ;
      for (i = 0; i < 4; i++)
      {
         VDD_VMMem_SelectPlane( i, 0 ) ;

         if ((pMS -> MSS_planes[ i ].MSS_plane_size != 0) || (i == 1))
         {
            if (pVDD_RestoreGraphics == (DWORD) VDD_VMMem_ZeroPage)
               VDD_VMMem_ZeroPage( uPhysPage ) ;
            else
               VDD_VMMem_CopyToPlane( hVM, &pMS -> MSS_planes[ i ],
                                      hPage, uPhysPage ) ;
         }
         else
         {
            // we are probably restoring a text page while the VM is
            // temporarily in a planar mode while it is loading a
            // soft font, so we need to restore plane 1 as a copy
            // of plane 0 shifted 1 byte.

            if (0 == (pVMCB -> VDD_Flags & (fVDD_Font | fVDD_DirFont)))
            {
               DEBUGMSG( "restoring plane 1 with no save mem and not loading font!\r\n" ) ;

               if (pVDD_RestoreGraphics == (DWORD) VDD_VMMem_ZeroPage)
                  VDD_VMMem_ZeroPage( uPhysPage ) ;
               else
                  VDD_VMMem_CopyToPlane( hVM, &pMS -> MSS_planes[ i ],
                                       hPage, uPhysPage ) ;
            }
            else
               VDD_VMMem_RestoreAttributePage( hVM, &pMS -> MSS_planes[ 0 ],
                                               hPage, uPhysPage ) ;
         }
      }

      QUEUEMSG( "VDD: end restore page...\r\n" ) ;

      vmmEndCriticalSection( hVM ) ;
      return TRUE ;
   }
   else
      goto RP_OddEven ;

} // VDD_VMMem_RestorePage()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_SelectPlane( BYTE bPlaneNum, BYTE bPlaneWMask )
//
//  Description:
//     Set read & write of specified plane.
//
//  Parameters:
//     BYTE bPlaneNum
//        plane # (0, 1, 2 or 3)
//
//     BYTE bPlaneWMask
//        plane write mask for Seq.2
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_SelectPlane
(
    BYTE            bPlaneNum,
    BYTE            bPlaneWMask
)
{
   volatile int  iDelay = 0 ;

   outp( pGrpIndx, 4 ) ;
   iDelay++ ;
   iDelay-- ;
   outp( pGrpIndx + 1, bPlaneNum ) ;
   iDelay++ ;
   iDelay-- ;
   outp( pSeqIndx, 2 ) ;
   iDelay++ ;
   iDelay-- ;
   outp( pSeqIndx + 1, bPlaneWMask ) ;

} // end of VDD_VMMem_SelectPlane()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_ChkPlaneSize( DWORD hVM, PMSSPLANEINFO pMPI,
//                               UINT uPgIndx )
//
//  Description:
//     Determine if page is part of allocated VM's save memory
//     for a particular plane.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PMSSPLANEINFO pMPI
//        pointer to MMSPLANEINFO structure
//
//     UINT uPgIndx
//        physical page index
//
//  Return Value:
//     BOOL
//        FALSE, if save memory doesn't include page
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_ChkPlaneSize
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    UINT            uPgIndx
)
{
   // Is page alloc'ed in plane?

   if (uPgIndx < pMPI -> MSS_plane_alloced)
      return ( TRUE ) ;

   // Is page within valid size of plane?

   if (uPgIndx >= pMPI -> MSS_plane_size)
      return ( FALSE ) ;

   if (vmmTestSysVMHandle( hVM ))
      VDD_Error_NoMainMem() ;

   return ( FALSE ) ;

} // end of VDD_VMMem_ChkPlaneSize()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_LockSaveMemPage( DWORD hVM, PMSSPLANEINFO pMPI,
//                                  UINT uPgIndx )
//
//  Description:
//     Locks down the save memory.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PMSSPLANEINFO pMPI
//        pointer to MMSPLANEINFO structure
//
//     UINT uPgIndx
//        physical page index
//
//  Return Value:
//     BOOL
//       FALSE, if savemem page doesn't exit, or can't be locked
// 
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_LockSaveMemPage
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    UINT            uPgIndx
)
{
   if (VDD_VMMem_ChkPlaneSize( hVM, pMPI, uPgIndx ))
      return ( vmmPageLock( pMPI -> MSS_plane_handle, 1, uPgIndx, 0 ) ) ;
   return ( FALSE ) ;

} // end of VDD_VMMem_LockSaveMemPage()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_CopyFromPlane( DWORD hVM, PMSSPLANEINFO pMPI,
//                                DWORD hPage, UINT uPhysPgNum )
//
//  Description:
//     Copy a 4Kb page from physical video memory to a page of
//     VM save memory.  This routine just does a straight
//     copy, so it will copy 4K bytes from which ever
//     plane(s) is(are) selected.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PMSSPLANEINFO pMPI
//        pointer to MMSPLANEINFO structure
//
//     DWORD hPage
//        handle to page
//
//     UINT uPhysPgNum
//        physical page number
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_CopyFromPlane
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPgNum
)
{
   int   i ;
   PSTR  pDest, pSrc ;

   if (VDD_VMMem_LockSaveMemPage( hVM, pMPI,
                                  ((PPAGEDATA) hPage) -> PDS_virtual_id ))
   {
      pDest =
         (PSTR) ((((PPAGEDATA) hPage) -> PDS_virtual_id << 12) +
                  pMPI -> MSS_plane_addr) ;
      pSrc =
         (PSTR) ((uPhysPgNum << 12) - 0xA0000 + dwVidPhysA0000) ;
      for (i = 0; i < P_SIZE ; i++)
         *pDest++ = *pSrc++ ;
      vmmPageUnLock( pMPI -> MSS_plane_handle, 1, uPhysPgNum, 0 ) ;
   }

} // end of VDD_VMMem_CopyFromPlane()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_SaveAttributePage
//
//  Description:
//     Copy a 4Kb page of text attribute bytes (even bytes from
//     plane 1) into the odd bytes in plane 0 save memory.
//     This routine requires that the hardware be set up in
//     planar read mode.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PMSSPLANEINFO pMPI
//        pointer to MMSPLANEINFO structure
//
//     DWORD hPage
//        handle to page
//
//     UINT uPhysPgNum
//        physical page number
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_SaveAttributePage
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPgNum
)
{
   int   i ;
   PSTR  pDest, pSrc ;

   if (VDD_VMMem_LockSaveMemPage( hVM, pMPI,
                                  ((PPAGEDATA) hPage) -> PDS_virtual_id ))
   {
      pDest =
         (PSTR) ((((PPAGEDATA) hPage) -> PDS_virtual_id << 12) +
                  pMPI -> MSS_plane_addr) ;
      pSrc =
         (PSTR) ((uPhysPgNum << 12) - 0xA0000 + dwVidPhysA0000) ;
      for (i = 0; i < P_SIZE / 2; i++)
      {
         pDest++ ;            // skip character byte in savemem
         *pDest++ = *pSrc++ ;
         pSrc++ ;             // skip undefined byte
      }
      vmmPageUnLock( pMPI -> MSS_plane_handle, 1, uPhysPgNum, 0 ) ;
   }

} // end of VDD_VMMem_SaveAttributePage()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_RestoreAttributePage( DWORD hVM, PMSSPLANEINFO pMPI,
//                                       DWORD hPage, UINT uPhysPgNum )
//
//  Description:
//     Copy a 4KB page of a text mode plane 0 save memory   
//     to a physical video memory page.  This routine requires
//     that the hardware be set up for a planar write.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PMSSPLANEINFO pMPI
//        pointer to MMSPLANEINFO structure
//
//     DWORD hPage
//        handle to page
//
//     UINT uPhysPgNum
//        physical page number
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_RestoreAttributePage
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPgNum
)
{
   int   i ;
   PSTR  pDest, pSrc ;

   if (VDD_VMMem_LockSaveMemPage( hVM, pMPI,
                                  ((PPAGEDATA) hPage) -> PDS_virtual_id ))
   {
      pSrc =
         (PSTR) ((((PPAGEDATA) hPage) -> PDS_virtual_id << 12) +
                  pMPI -> MSS_plane_addr) ;

      // copy plane 0 to plane 1 shift 1 byte to left

      pSrc++ ;    
      pDest =
         (PSTR) ((uPhysPgNum << 12) - 0xA0000 + dwVidPhysA0000) ;

      // Ignore last byte

      for (i = 0; i < P_SIZE - 1; i++)
         *pDest++ = *pSrc++ ;
      vmmPageUnLock( pMPI -> MSS_plane_handle, 1, uPhysPgNum, 0 ) ;
   }
   else
      VDD_VMMem_InitText( uPhysPgNum, 0x2007 ) ;

} // end of VDD_VMMem_RestoreAttributePage()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_CopyToPlane( DWORD hVM, PMSSPLANEINFO pMPI,
//                              DWORD hPage, UINT uPhysPgNum )
//
//  Description:
//     Copy a 4Kb page from VM save memory to a physical video   
//     memory page.  This routine just does a straight
//     copy, so it will copy 4K bytes to which ever
//     plane(s) is(are) selected.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PMSSPLANEINFO pMPI
//        pointer to MMSPLANEINFO structure
//
//     DWORD hPage
//        handle to page
//
//     UINT uPhysPgNum
//        physical page number
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_CopyToPlane
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPgNum
)
{
   int   i ;
   PSTR  pDest, pSrc ;

   if (VDD_VMMem_LockSaveMemPage( hVM, pMPI,
                                  ((PPAGEDATA) hPage) -> PDS_virtual_id ))
   {
      pSrc =
         (PSTR) ((((PPAGEDATA) hPage) -> PDS_virtual_id << 12) +
                  pMPI -> MSS_plane_addr) ;
      pDest =
         (PSTR) ((uPhysPgNum << 12) - 0xA0000 + dwVidPhysA0000) ;
      for (i = 0; i < P_SIZE ; i++)
         *pDest++ = *pSrc++ ;
      vmmPageUnLock( pMPI -> MSS_plane_handle, 1, uPhysPgNum, 0 ) ;
   }
   else
      VDD_VMMem_ZeroPage( uPhysPgNum ) ;

} // end of VDD_VMMem_CopyToPlane()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_ZeroPage( UINT uPhysPgNum )
//
//  Description:
//     Default restore routine which will write a 4Kb page with
//     0's, if VDD_VMMem_RestorePage is called for a page
//     which has never been saved.
//
//  Parameters:
//     UINT uPhysPgNum
//        physical page number
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_ZeroPage
(
    UINT            uPhysPgNum
)
{
   int   i ;
   PSTR  pDest ;

   pDest =
      (PSTR) ((uPhysPgNum << 12) - 0xA0000 + dwVidPhysA0000) ;
   for (i = 0; i < P_SIZE ; i++)
      *pDest++ = 0 ;

} // end of VDD_VMMem_ZeroPage()

//--------------------------------------------------------------------------
//  
//  VOID VDD_VMMem_InitText
//  
//  Description:
//      Default restore routine which will write a 4Kb page with
//      [' ', 7]'s (a character space and an attribute of 7), if
//      VDD_VM_Mem_Restore_Page is called for a text page which
//      has never been saved.
//  
//  Parameters:
//      UINT uPhysPage
//
//      WORD wValue
//  
//  Return (VOID):
//  
//  History:   Date       Author      Comment
//             10/23/93   BryanW      
//  
//--------------------------------------------------------------------------

VOID VDD_VMMem_InitText
(
    UINT            uPhysPage,
    WORD            wValue
    
)
{
   UINT   i ;
   PWORD  pBuf ;

   uPhysPage <<= 12 ;
   uPhysPage -= 0xA0000 ;
   uPhysPage += dwVidPhysA0000 ;
   pBuf = (PWORD) uPhysPage ;

   for (i = 0; i < P_SIZE / 2; i++)
      *pBuf++ = wValue ;

} // VDD_VMMem_InitText()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_AllocMsgModeMem( VOID )
//
//  Description:
//     Allocate the visible pages of video memory needed for
//     message mode for VDD use.
//
//     By allocating them from VDDPHMem we force the current
//     memory to be saved for the current owner.
//
//  Parameters:
//     Nothing.
//
//  History:   Date       Author      Comment
//              4/27/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_AllocMsgModeMem()
{
   int     i ;
   PVDDCB  pVMCB ;
   
   pVMCB = (PVDDCB) (hVidMsgPseudoVM + dwVidCBOff) ;

   VDD_VMMem_MemC_AboutToChange( hVidMsgPseudoVM, pVMCB ) ;

   for (i = 0; i < bVidMsgPages; i++)
      VDD_PHMem_AllocVideoPage( hVidMsgPseudoVM, pVMCB, i,
                                VDD_VMMem_PageHandler ) ;

} // end of VDD_VMMem_AllocMsgModeMem()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_SaveDirtyState( DWORD hVM, PVDDCB pVMCB )
//
//  Description:
//     Walk thru list of all video pages owned by a VM; OR the   
//     physical dirty bits from the page table of any mapped
//     page into the page flags for that page and clear the
//     physical dirty bit.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//  History:   Date       Author      Comment
//              5/31/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_SaveDirtyState
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   int    i ;
   BYTE   bTemp ;
   DWORD  dwLinPage, hPage ;
   PBYTE  ptr ;

   for (i = 0; i < pVMCB -> VDD_VideoPages; i++)
   {
      if ((hPage = pVMCB -> VDD_PageMap[ i ]) &&
          (dwLinPage = ((PPAGEDATA) hPage) -> PDS_linear_page))
      {
         if (VDD_VMMem_GetPTE( hVM, dwLinPage ) & P_DIRTY)
         {
            ((PPAGEDATA) hPage) -> PDS_flags |= fPDS_Dirty ;
            vmmModifyPageBits( hVM, dwLinPage, 1, -1, 0, PG_IGNORE, 0 ) ;

            // Touch page to set its accessed bit so that we don't
            // think we are in a dead lock situation at page fault time.

            ptr =
               (PBYTE)(dwLinPage << 12 + ((PVMMCB) hVM) -> CB_High_Linear) ;
            bTemp = *ptr ;

            VDD_VMMem_RecordVRAMMod( pVMCB, i ) ;
         }
      }
   }

} // end of VDD_VMMem_SaveDirtyState()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_FreePages( DWORD hVM, PVDDCB pVMCB )
//
//  Description:
//     Free all video pages owned by VM.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//  History:   Date       Author      Comment
//              4/27/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_FreePages
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   int    i ;
   DWORD  hPage ;

   if (pVMCB -> VDD_Owned_Pages)
   {
      VDD_VMMem_Disable( hVM, pVMCB ) ;

      for (i = 0; i < pVMCB -> VDD_VideoPages; i++)
      {
         hPage = pVMCB -> VDD_PageMap[ i ] ;
         if (hPage)
         {
#ifdef DEBUG
            // Verify that all phys pages have been unmapped
            // by calling VDD_VMMem_IsPageDirty.  VDD_VMMem_Disable,
            // called above, should have unmapped all pages and called
            // VDD_VMMem_IsPageDirty in the process so that the
            // physical dirty bit was OR'ed into PDS_flags.

            if (VDD_VMMem_IsPageDirty( hVM, hPage ))
               DEBUGMSG( "Phys page still mapped at end of VDD_VMMem_FreePages\r\n" ) ;
#endif

            // Don't need to save the page...

            ((PPAGEDATA) hPage) -> PDS_flags &= ~fPDS_Dirty ;

            VDD_PHMem_FreePage( hPage ) ;
         }
      }
   }

} // end of VDD_VMMem_FreePages()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_RecordVRAMMod( PVDDCB pVMCB, UINT uPageID )
//
//  Description:
//     Record a video page modification (indicated by the dirty
//     bit being set) by setting the VM flag in VDD_Flags and
//     by setting the page bit in the modified pages array
//     (VDD_VRAM_Mods).
//
//  Parameters:
//     PVDDCB pVMCB
//        pointer to VM's control block
//
//     UINT uPageID
//        virtual page ID
//
//  History:   Date       Author      Comment
//              4/24/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_RecordVRAMMod
(
    PVDDCB          pVMCB,
    UINT            uPageID
)
{
   pVMCB -> VDD_ModTime = vmmGetLastUpdatedSystemTime() ;
   pVMCB -> VDD_VRAM_Mods[ 0 ] |= (1 << uPageID) ;

} // end of VDD_VMMem_RecordVRAMMod()


#ifdef DEBUG

//------------------------------------------------------------------------
//  VOID VDD_VMMem_DebugDumpPages( DWORD hVM, PVDDCB pVMCB )
//
//  Description:
//     Dumps a list of tables show VM memory usage for debugging.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//  History:   Date       Author      Comment
//              5/31/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_DebugDumpPages
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   TRACEMSGPARAM( "display virtual video page table, VM owns #EAX phys pages\r\n",
                  pVMCB -> VDD_Owned_Pages ) ;

} // VDD_VMMem_DebugDumpPages

//---------------------------------------------------------------------------
//  End of File: vddvmmem.c
//---------------------------------------------------------------------------

