//---------------------------------------------------------------------------
//
//  Module: vglobald.c
//
//  Purpose:
//     Allocates global V86 memory for use with "Windows-aware" TSR.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              6/10/93   BryanW      Wrote it.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vglobald.h"

#pragma data_seg( "_LDATA", "CODE" ) ;

//=============================================================================
//          V I R T U A L   D E V I C E   D E C L A R A T I O N
//=============================================================================

DDB VGLOBALD_DDB = { NULL,                  // reserved, must be NULL
                     DDK_Version,           // reserved, DDK_Version
                     VGLOBALD_Device_ID,    // Device ID
                     VGLOBALD_Major_Ver,    // Major Version
                     VGLOBALD_Minor_Ver,    // Minor Version
                     NULL,
                     "VGLOBALD",
                     Undefined_Init_Order,
                     (DWORD) vmmwrapVxDControlProc,
                     NULL,
                     NULL,
                     NULL,
                     NULL,
                     NULL,
                     NULL,
                     NULL } ;

//
// This table is used by the vmmwrapVxDControlProc defined in VMMWRAP.ASM.                  
// It lists the messages and associated dispatch functions, it
// must be terminated with -1 and NULL.                  
//

DISPATCHINFO alpVxDDispatchProcs[] =
   { VM_Init,             VGLOBALD_VM_Init,
     Destroy_VM,          VGLOBALD_Destroy_VM,
     Device_Init,         VGLOBALD_Device_Init,
     -1,                  NULL } ;


// globals

UINT     guCBOffset ;                // VM control block offset
DWORD    ghOwnerVM = NULL ;          // Data area owner
WORD     gwGlobalPage = NULL ;       // Global V86 data area
PDWORD   gpPageDir ;                 // Page directory pointer (linear addr)
IPFPROC  pVGLOBALD_Invalid_PFault ;  // Invalid page fault procedure thunk

//------------------------------------------------------------------------
//  BOOL VGLOBALD_VM_Init
//
//  Description:
//     Modify the global page's PTE for this VM to make the page
//     not present or available.
//
//  Parameters:
//     hVM
//        VM handle
//
//  History:   Date       Author      Comment
//              6/13/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL CDECL VGLOBALD_VM_Init
(
   DWORD    hVM
)
{
   PVMCB   pVMCB ;

   vmmTraceOut( "VGLOBALD_VM_Init\r\n" ) ;

   pVMCB = (PVMCB) (hVM + guCBOffset) ;

   // An assumption is made that this page is within
   // the 1st MB because it is V86 addressable.

   pVMCB -> pPageTable =
      (PDWORD) vmmMapPhysToLinear( (DWORD)(gpPageDir[ 0 ] & 0xFFFFF000),
                                   P_SIZE, NULL ) ;

   // Modify PTE to force it not present or available

   pVMCB -> pPageTable[ gwGlobalPage ] &= ~P_AVAIL ;

   return ( TRUE ) ;
   
} // end of VGLOBALD_VM_Init()

//------------------------------------------------------------------------
//  BOOL VGLOBALD_Destroy_VM
//
//  Description:
//     If owning VM, sets owner to NULL.
//
//  Parameters:
//     hVM
//        VM handle
//
//  History:   Date       Author      Comment
//              6/13/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL CDECL VGLOBALD_Destroy_VM
(
   DWORD    hVM
)
{
   vmmTraceOut( "VGLOBALD_Destroy_VM\r\n" ) ;

   if (ghOwnerVM == hVM)
      ghOwnerVM = NULL ;

   return ( TRUE ) ;
   
} // end of VGLOBALD_Destroy_VM()

//------------------------------------------------------------------------
//  BOOL VGLOBALD_Invalid_PFault
//
//  Description:
//     Resolves the page fault by mapping memory allocated
//     during VM initialization.
//
//  Parameters:
//     DWORD hVM
//        handle to VM causing fault
//
//     PIPFDATA pIPFData
//        pointer to invalid page fault data
//
//  History:   Date       Author      Comment
//              6/13/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL CDECL VGLOBALD_Invalid_PFault
(
   DWORD    hVM,
   PIPFDATA pIPFData
)
{
   PVMCB   pVMCB, pOwnerVMCB ;

   vmmTraceOutParam( "Page fault for PTE:#EAX\r\n",
                     pIPFData -> IPF_PTEEntry ) ;

   pVMCB = (PVMCB) (hVM + guCBOffset) ;

   if (pVMCB -> pPageTable[ gwGlobalPage ]  != pIPFData -> IPF_PTEEntry)
      return ( FALSE ) ;

   vmmBeginCriticalSection( hVM,
                            Block_Svc_If_Ints_Locked | Block_Enable_Ints ) ;

   if (ghOwnerVM)
   {
      pOwnerVMCB = (PVMCB) (ghOwnerVM + guCBOffset) ;
      pOwnerVMCB -> pPageTable[ gwGlobalPage ] &= ~P_AVAIL ;
   }

   pVMCB -> pPageTable[ gwGlobalPage ] |= P_AVAIL ;
   ghOwnerVM = hVM ;

   vmmEndCriticalSection( hVM ) ;

   return ( TRUE ) ;

} // end of VGLOBALD_Invalid_PFault()

//---------------------------------------------------------------------------
//  End of File: vglobald.c
//---------------------------------------------------------------------------
